# Copyright (C) 2023. BMW Car IT GmbH. All rights reserved.
import logging
from multiprocessing import Event, Queue
import os
import time
import tempfile
import unittest
from queue import Empty

from dlt.dlt_broker_handlers import DLTFileSpinner
from tests.utils import (
    create_messages,
    stream_multiple,
    stream_with_params,
    append_stream_to_file,
    append_message_to_file,
)

logger = logging.getLogger(__name__)


class TestDLTFileSpinner(unittest.TestCase):
    def setUp(self):
        self.filter_queue = Queue()
        self.message_queue = Queue()
        self.stop_event = Event()
        # Dlt file is created with empty content
        _, self.dlt_file_name = tempfile.mkstemp(suffix=b".dlt")
        self.dlt_file_spinner = DLTFileSpinner(
            self.filter_queue, self.message_queue, self.stop_event, self.dlt_file_name
        )
        # dispatched_messages from DLTFileSpinner.message_queue
        self.dispatched_messages = []

    def tearDown(self):
        if self.dlt_file_spinner.is_alive():
            self.dlt_file_spinner.break_blocking_main_loop()
            self.stop_event.set()
            self.dlt_file_spinner.join()
        if os.path.exists(self.dlt_file_name):
            os.remove(self.dlt_file_name)

    def test_init(self):
        self.assertFalse(self.dlt_file_spinner.mp_stop_flag.is_set())
        self.assertFalse(self.dlt_file_spinner.is_alive())
        self.assertTrue(self.dlt_file_spinner.filter_queue.empty())
        self.assertTrue(self.dlt_file_spinner.message_queue.empty())

    def test_run_basic_without_dlt_file(self):
        # Delete the created dlt file
        os.remove(self.dlt_file_name)

        self.assertFalse(self.dlt_file_spinner.is_alive())
        self.dlt_file_spinner.start()
        self.assertTrue(self.dlt_file_spinner.is_alive())
        self.assertNotEqual(self.dlt_file_spinner.pid, os.getpid())
        # DLT file does NOT exist
        self.assertFalse(os.path.exists(self.dlt_file_spinner.file_name))

        self.dlt_file_spinner.break_blocking_main_loop()
        self.stop_event.set()
        self.dlt_file_spinner.join()
        self.assertFalse(self.dlt_file_spinner.is_alive())

    def test_run_basic_with_empty_dlt_file(self):
        self.assertFalse(self.dlt_file_spinner.is_alive())
        self.dlt_file_spinner.start()
        self.assertTrue(self.dlt_file_spinner.is_alive())
        self.assertNotEqual(self.dlt_file_spinner.pid, os.getpid())
        # dlt_reader is instantiated and keeps alive
        self.assertTrue(os.path.exists(self.dlt_file_spinner.file_name))
        # Expect no dlt log is dispatched
        time.sleep(2)
        self.assertTrue(self.dlt_file_spinner.message_queue.empty())
        # First stop dlt reader, then stop DLTFileSpinner
        self.dlt_file_spinner.break_blocking_main_loop()
        self.stop_event.set()
        self.dlt_file_spinner.join()
        self.assertFalse(self.dlt_file_spinner.is_alive())

    def test_handle_add_new_filter(self):
        self.dlt_file_spinner.filter_queue.put(("queue_id", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.dlt_file_spinner.handle(None)
        self.assertIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)
        self.assertEqual(self.dlt_file_spinner.context_map[("SYS", "JOUR")], ["queue_id"])

    def test_handle_remove_filter_single_entry(self):
        self.dlt_file_spinner.filter_queue.put(("queue_id", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.dlt_file_spinner.handle(None)
        self.assertIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)
        self.assertEqual(self.dlt_file_spinner.context_map[("SYS", "JOUR")], ["queue_id"])

        self.dlt_file_spinner.filter_queue.put(("queue_id", [("SYS", "JOUR")], False))
        time.sleep(0.01)
        self.dlt_file_spinner.handle(None)
        self.assertNotIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)

    def test_handle_remove_filter_multiple_entries(self):
        self.dlt_file_spinner.filter_queue.put(("queue_id1", [("SYS", "JOUR")], True))
        self.dlt_file_spinner.filter_queue.put(("queue_id2", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.dlt_file_spinner.handle(None)
        self.assertIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)
        self.assertEqual(self.dlt_file_spinner.context_map[("SYS", "JOUR")], ["queue_id1", "queue_id2"])

        self.dlt_file_spinner.filter_queue.put(("queue_id1", [("SYS", "JOUR")], False))
        time.sleep(0.01)
        self.dlt_file_spinner.handle(None)
        self.assertIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)
        self.assertEqual(self.dlt_file_spinner.context_map[("SYS", "JOUR")], ["queue_id2"])

    def test_handle_multiple_similar_filters(self):
        self.dlt_file_spinner.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.dlt_file_spinner.filter_queue.put(("queue_id1", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.dlt_file_spinner.handle(None)
        self.assertIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)
        self.assertEqual(self.dlt_file_spinner.context_map[("SYS", "JOUR")], ["queue_id0", "queue_id1"])

    def test_handle_multiple_different_filters(self):
        self.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.filter_queue.put(("queue_id1", [("DA1", "DC1")], True))
        time.sleep(0.01)
        self.dlt_file_spinner.handle(None)
        self.assertIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)
        self.assertIn(("DA1", "DC1"), self.dlt_file_spinner.context_map)
        self.assertEqual(self.dlt_file_spinner.context_map[("SYS", "JOUR")], ["queue_id0"])
        self.assertEqual(self.dlt_file_spinner.context_map[("DA1", "DC1")], ["queue_id1"])

    def test_handle_message_tag_and_distribute(self):
        self.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.filter_queue.put(("queue_id1", [("DA1", "DC1")], True))
        self.filter_queue.put(("queue_id2", [("SYS", None)], True))
        self.filter_queue.put(("queue_id3", [(None, "DC1")], True))
        self.filter_queue.put(("queue_id4", [(None, None)], True))
        time.sleep(0.01)

        # - simulate receiving of messages
        for _ in range(10):
            for message in create_messages(stream_multiple, from_file=True):
                self.dlt_file_spinner.handle(message)

        self.assertIn(("SYS", "JOUR"), self.dlt_file_spinner.context_map)
        self.assertIn(("DA1", "DC1"), self.dlt_file_spinner.context_map)
        self.assertIn((None, None), self.dlt_file_spinner.context_map)
        self.assertIn(("SYS", None), self.dlt_file_spinner.context_map)
        self.assertIn((None, "DC1"), self.dlt_file_spinner.context_map)
        try:
            # 60 == 10 messages of each for SYS, JOUR and None combinations +
            #       10 for (None,None)
            messages = [self.message_queue.get(timeout=0.01) for _ in range(60)]

            # these queues should not get any messages from other queues
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id0"]), 10)
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id1"]), 10)
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id2"]), 10)
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id3"]), 10)
            # this queue should get all messages
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id4"]), 20)
        except Empty:
            # - we should not get an Empty for at least 40 messages
            self.fail()

    def _update_dispatch_messages_from_dlt_file_spinner(self):
        for index in range(60):
            try:
                message = self.dlt_file_spinner.message_queue.get(timeout=0.01)
                if not self.dispatched_messages or message[1] != self.dispatched_messages[-1][1]:
                    self.dispatched_messages.append(message)
            except:  # noqa: E722
                pass

    def test_run_with_writing_to_file(self):
        """
        Test with real dlt file, which is written at runtime

        1. set filter_queue properly, so that the handled messages could be added to message_queue later
        2. start DLTFileSpinner
           At this moment, no messages are written to dlt file, so no messages in DLTFileSpinner.message_queue
        3. write 2 sample messages to dlt file
           Expectation: we could dispatch 2 messages from DLTFileSpinner.message_queue
        5. stop DLTFileSpinner
        """
        # 1. set filter_queue properly, so that the handled messages could be added to message_queue later
        self.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.filter_queue.put(("queue_id1", [("DA1", "DC1")], True))
        self.filter_queue.put(("queue_id2", [("SYS", None)], True))
        self.filter_queue.put(("queue_id3", [(None, "DC1")], True))
        self.filter_queue.put(("queue_id4", [(None, None)], True))
        time.sleep(0.01)
        # 2. start DLTFileSpinner
        self.assertFalse(self.dlt_file_spinner.is_alive())
        self.dlt_file_spinner.start()
        self.assertTrue(self.dlt_file_spinner.is_alive())
        self.assertNotEqual(self.dlt_file_spinner.pid, os.getpid())
        # dlt_reader is instantiated and keeps alive
        self.assertTrue(os.path.exists(self.dlt_file_spinner.file_name))
        # With empty file content, no messages are dispatched to message_queue
        time.sleep(2)
        self.assertTrue(self.dlt_file_spinner.message_queue.empty())
        # 3. write 2 sample messages to dlt file
        append_stream_to_file(stream_multiple, self.dlt_file_name)
        # Expect the written dlt logs are dispatched to message_queue
        self._update_dispatch_messages_from_dlt_file_spinner()
        self.assertEqual(2, len(self.dispatched_messages))
        # 4. stop DLTFileSpinner
        self.dlt_file_spinner.break_blocking_main_loop()
        self.stop_event.set()
        self.dlt_file_spinner.join()
        self.assertFalse(self.dlt_file_spinner.is_alive())

    def test_run_with_writing_to_file_twice(self):
        """
        Test with real dlt file, which is written at runtime 2 times

        1. set filter_queue properly, so that the handled messages could be added to message_queue later
        2. start DLTFileSpinner
        3. write 2 sample messages to dlt file
           Expectation: we could dispatch 2 messages from DLTFileSpinner.message_queue
        4. append 1 sample message to dlt file
           Expectation: we could dispatch 3 messages from DLTFileSpinner.message_queue
        5. stop DLTFileSpinner
        """
        # 1. set filter_queue properly, so that the handled messages could be added to message_queue later
        self.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.filter_queue.put(("queue_id1", [("DA1", "DC1")], True))
        self.filter_queue.put(("queue_id2", [("SYS", None)], True))
        self.filter_queue.put(("queue_id3", [(None, "DC1")], True))
        self.filter_queue.put(("queue_id4", [(None, None)], True))
        time.sleep(0.01)
        # 2. start DLTFileSpinner
        self.assertFalse(self.dlt_file_spinner.is_alive())
        self.dlt_file_spinner.start()
        self.assertTrue(self.dlt_file_spinner.is_alive())
        self.assertNotEqual(self.dlt_file_spinner.pid, os.getpid())
        # dlt_reader is instantiated and keeps alive
        self.assertTrue(os.path.exists(self.dlt_file_spinner.file_name))
        # With empty file content, no messages are dispatched to message_queue
        time.sleep(2)
        self.assertTrue(self.dlt_file_spinner.message_queue.empty())
        # 3. write 2 sample messages to dlt file
        append_stream_to_file(stream_multiple, self.dlt_file_name)
        # Expect the written dlt logs are dispatched to message_queue
        self._update_dispatch_messages_from_dlt_file_spinner()
        self.assertEqual(2, len(self.dispatched_messages))
        # 4. append 1 sample message to dlt file
        append_stream_to_file(stream_with_params, self.dlt_file_name)
        self._update_dispatch_messages_from_dlt_file_spinner()
        self.assertEqual(3, len(self.dispatched_messages))
        # 5. stop DLTFileSpinner
        self.dlt_file_spinner.break_blocking_main_loop()
        self.stop_event.set()
        self.dlt_file_spinner.join()
        self.assertFalse(self.dlt_file_spinner.is_alive())

    def test_run_with_writing_empty_apid_ctid_to_file(self):
        """
        Test with real dlt file, which contains message with apid=b"" and ctid=b""

        1. set filter_queue properly, so that the handled messages could be added to message_queue later
        2. start DLTFileSpinner
           At this moment, no messages are written to dlt file, so no messages in DLTFileSpinner.message_queue
        3. write message with apid=b"" and ctid=b"" to dlt file
           Expectation: we could dispatch 1 message from DLTFileSpinner.message_queue
                        and, apid==b"" and ctid==b""
        5. stop DLTFileSpinner
        """
        # 1. set filter_queue properly, so that the handled messages could be added to message_queue later
        self.filter_queue.put(("queue_id0", [(None, None)], True))
        time.sleep(0.01)
        # 2. start DLTFileSpinner
        self.assertFalse(self.dlt_file_spinner.is_alive())
        self.dlt_file_spinner.start()
        self.assertTrue(self.dlt_file_spinner.is_alive())
        self.assertNotEqual(self.dlt_file_spinner.pid, os.getpid())
        # dlt_reader is instantiated and keeps alive
        self.assertTrue(os.path.exists(self.dlt_file_spinner.file_name))
        # With empty file content, no messages are dispatched to message_queue
        time.sleep(2)
        self.assertTrue(self.dlt_file_spinner.message_queue.empty())
        # 3. write a message to dlt file
        # Construct a message with apid==b"" and ctid==b""
        message = create_messages(stream_with_params, from_file=True)[0]
        message.extendedheader.apid = b""
        message.extendedheader.ctid = b""
        # Write this message into dlt file
        append_message_to_file(message, self.dlt_file_name)
        # Expect the written dlt logs are dispatched to message_queue
        self._update_dispatch_messages_from_dlt_file_spinner()
        self.assertEqual(1, len(self.dispatched_messages))
        # Expectation: the received message should have apid==b"" and ctid==b""
        self.assertEqual("", self.dispatched_messages[0][1].apid)
        self.assertEqual("", self.dispatched_messages[0][1].ctid)
        # 4. stop DLTFileSpinner
        self.dlt_file_spinner.break_blocking_main_loop()
        self.stop_event.set()
        self.dlt_file_spinner.join()
        self.assertFalse(self.dlt_file_spinner.is_alive())



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.
"""Test helpers and data"""

import atexit
import ctypes
import io
import tempfile
import os

from dlt.dlt import DLTClient, load

stream_one = io.BytesIO(b"5\x00\x00 MGHS\xdd\xf6e\xca&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00")

stream_with_params = (
    b"DLT\x01\xc2<\x85W\xc7\xc5\x02\x00MGHS=r\x00\xa0MGHS\x00\x00\x02B\x00X\xd4\xf1A\x08"
    b"ENV\x00LVLM\x00\x02\x00\x00-\x00CLevelMonitor::notification() => commandType\x00#"
    b"\x00\x00\x00\x03\x00\x00\x00\x00\x02\x00\x00\t\x00deviceId\x00#\x00\x00\x00\x05\x00"
    b"\x00\x00\x00\x02\x00\x00\x06\x00value\x00#\x00\x00\x00\xea\x0f\x00\x00\x00\x02\x00"
    b"\x00\x12\x00simulation status\x00#\x00\x00\x00\x00\x00\x00\x00"
)

stream_multiple = (
    b"DLT\x01#o\xd1WD>\x0c\x00MGHS5\x00\x00YMGHS\x00\x01\x80\xd1&\x01DA1\x00DC1\x00\x03\x00\x00\x00"
    b"\x07\x01\x00SYS\x00\x01\x00FILE\xff\xff\x16\x00File transfer manager.\x12\x00"
    b"DLT System ManagerremoDLT\x01#o\xd1Wo>\x0c\x00MGHS=\x00\x01PMGHS\x00\x00\x03\xf4\x00"
    b"\x01i\xa6A\x05SYS\x00JOUR\x00\x02\x00\x00\x1b\x002011/11/11 11:11:18.005274\x00\x00\x02\x00\x00"
    b"\t\x006.005274\x00\x00\x02\x00\x00\x16\x00systemd-journal[748]:\x00\x00\x02\x00\x00\x0f\x00"
    b"Informational:\x00\x00\x02\x00\x00\xcf\x00Runtime journal (/run/log/journal/) is currently"
    b" using 8.0M.\nMaximum allowed usage is set to 385.9M.\nLeaving at least 578.8M free (of"
    b" currently available 3.7G of space).\nEnforced usage limit is thus 385.9M.\x00"
)

stream_multiple_with_malformed_message_at_begining = (
    b"DLT\x01\xfar\xc5c\xf7j\x03\x00\x00\x00\x00\x00\x00\x00\x00LCMFLOW WUP invalidDLT\x01"
    b"\xfar\xc5c\x0bo\x03\x00XORA'\x01\x00\x1bXORA\x16\x02\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x11\x04\x00\x00\x00\x00DLT\x01\xfar\xc5c?o\x03"
    b"DLT\x01#o\xd1WD>\x0c\x00MGHS5\x00\x00YMGHS\x00\x01\x80\xd1&\x01DA1\x00DC1\x00\x03\x00\x00\x00"
    b"\x07\x01\x00SYS\x00\x01\x00FILE\xff\xff\x16\x00File transfer manager.\x12\x00"
    b"DLT System ManagerremoDLT\x01#o\xd1Wo>\x0c\x00MGHS=\x00\x01PMGHS\x00\x00\x03\xf4\x00"
    b"\x01i\xa6A\x05SYS\x00JOUR\x00\x02\x00\x00\x1b\x002011/11/11 11:11:18.005274\x00\x00\x02\x00\x00"
    b"\t\x006.005274\x00\x00\x02\x00\x00\x16\x00systemd-journal[748]:\x00\x00\x02\x00\x00\x0f\x00"
    b"Informational:\x00\x00\x02\x00\x00\xcf\x00Runtime journal (/run/log/journal/) is currently"
    b" using 8.0M.\nMaximum allowed usage is set to 385.9M.\nLeaving at least 578.8M free (of"
    b" currently available 3.7G of space).\nEnforced usage limit is thus 385.9M.\x00"
)

msg_benoit = (
    b"DLT\x01\xa5\xd1\xceW\x90\xb9\r\x00MGHS=\x00\x00RMGHS\x00\x00\n[\x00\x0f\x9b#A\x01DEMODATA\x00"
    b"\x82\x00\x002\x00Logging from the constructor of a global instance\x00"
)


control_one = (
    b"DLT\x01#o\xd1W\x99!\x0c\x00MGHS5\x00\x00;MGHS\x00\x01\x7f\xdb&\x01DA1\x00DC1\x00\x03"
    b"\x00\x00\x00\x07\x01\x00HDDM\x01\x00CAPI\xff\xff\x04\x00CAPI\x06\x00hddmgrremo"
)

# DLT file with invalid storage header and frames
file_storage_clean = (
    b"DLT\x01\x9a\xc6\xbfW\x020\t\x00MGHS5\x00\x00 MGHS\x00\x02\x8aC&\x01DA1\x00DC1"
    b"\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00DLT\x01\x9a\xc6\xbfWoA\t\x00MGHS="
    b"\x00\x00NMGHS\x00\x00\x049\x00\x01p<A\x01DLTDINTM\x00\x02\x00\x00.\x00Daemon "
    b"launched. Starting to output traces...\x00DLT\x01\x9a\xc6\xbfW_H\t\x00MGHS=\x01"
    b"\x00qMGHS\x00\x00\x049\x00\x01pxA\x01DLTDINTM\x00\x02\x00\x00Q\x00ApplicationID"
    b"'PERD' registered for PID 987, Description=Personalization Daemon\n\x00"
)

# DLT file with invalid storage header but valid frames
file_storage_invalid_storage_hdr = (
    b"DLT\x01V\x03EX8\x06\x0b\x00\x00\x00\x00\x00=\x9d\x00\xaaMGHS\x00"
    b"\x00\x04$\x00\n\x11\xed1\x01NAVCSPCODLT\x01V\x03EXuD\x0c\x00\x00"
    b"\x00\x00\x00'\x01\x00\x1bXORA\x16\x02\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x11\x04\x00\x00\x00\x00DLT\x01V\x03EX\xfd[\x0c\x00"
    b"\x00\x00\x00\x005\x00\x00 MGHS\x00\n\x14\xda&\x01DA1\x00DC1\x00\x02"
    b"\x0f\x00\x00\x00\x02\x00\x00\x00\x00DLT\x01V\x03EXel\x0c\x00\x00\x00"
    b"\x00\x005\x00\x00 MGHS\x00\n\x11\xfb&\x01DA1\x00DC1\x00\x02\x0f\x00"
    b"\x00\x00\x01\x00\x00\x00\x00DLT\x01V\x03EXvl\x0c\x00\x00\x00\x00\x00"
    b"=\x86\x00cMGHS\x00\x00\x07\xa7\x00\n\x12\x1cA\x01NAVIASIA\x00\x02\x00"
    b"\x00C\x00[MSG_NV_MD] [NVS_MV][WIN1]iparam_update gui_mode(0) shmem "
    b"val(0) \n\x00DLT\x01V\x03EX}l\x0c\x00\x00\x00\x00\x00=\x87\x00WMGHS"
    b"\x00\x00\x07\xa7\x00\n\x12!A\x01NAVIASIA\x00\x02\x00\x007\x00[MSG_NV"
    b"_MD] [NVS_MV][WIN1] full menu stratframe(1479)\n\x00DLT\x01V\x03EX\x86"
    b"l\x0c\x00\x00\x00\x00\x00=n\x00\x89MGHS\x00\x00\x04$\x00\n\x12*A\x01NAV"
    b"CBLPS\x00\x02\x00\x00i\x00[Thread A] [PositioningCanAdapter.cpp::operat"
    b"or():92] Send: GPSRo"
)

file_with_four_lifecycles = (
    b"DLT\x01\xc5\x82\xdaX\x19\x93\r\x00XORA'\x01\x00\x1bXORA"  # trace to buffer
    b"\x16\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11\x04\x00\x00\x00\x00"
    b"DLT\x01\xc5\x82\xdaXQi\x0e\x00MGHS5\x00\x00 MGHS"  # trace to buffer
    b"\x00\x03U\xe0&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
    b"DLT\x01\xc5\x82\xdaX\x82o\x0e\x00MGHS=\x00\x00NMGHS"  # first lifecycle
    b"\x00\x00\x02r\x00\x00\x8frA\x01DLTDINTM\x00\x02\x00\x00.\x00"
    b"Daemon launched. Starting to output traces...\x00"
    b"DLT\x01\xc9\xc1\x91Y\xbf\x1b\x00\x00MGHS5\x00\x00 MGHS"  # trace to buffer
    b"\x00\x00v\n&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
    b"DLT\x01\xc9\xc1\x91Y\x9f/\x00\x00MGHS=\x00\x00NMGHS"  # new lifecycle
    b"\x00\x00\x032\x00\x00IWA\x01DLTDINTM\x00\x02\x00\x00.\x00"
    b"Daemon launched. Starting to output traces...\x00"
    b"DLT\x01m\xc2\x91Y\x9f\xda\x07\x00MGHS5\x00\x00 MGHS"  # no new lifecycle
    b"\x00\x00_\xde&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
    b"DLT\x01m\xc2\x91Y\xad\xe4\x07\x00MGHS=\x01\x00zMGHS"  # random trace
    b"\x00\x00\x02\xab\x00\x00@VA\x01DLTDINTM\x00\x02\x00\x00Z\x00"
    b"ApplicationID 'DBSY' registered for PID 689, Description=DBus"
    b" Logging|SysInfra|Log&Trace\n\x00"
    b"DLT\x01\xed\xc2\x91Y\x0f\xf0\x08\x00MGHS5\x00\x00 MGHS"  # trace to buffer
    b"\x00\x00\x9dC&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
    b"DLT\x01\xed\xc2\x91Y\x17.\n\x00MGHS=\x00\x00NMGHS"  # new lifecycle
    b"\x00\x00\x02\xae\x00\x00@/A\x01DLTDINTM\x00\x02\x00\x00.\x00"
    b"Daemon launched. Starting to output traces...\x00"
    b"DLT\x01]\xc3\x91Y,\x91\r\x00MGHS=\x00\x00NMGHS"  # new lifecycle
    b"\x00\x00\x02\xbd\x00\x00G\xefA\x01DLTDINTM\x00\x02\x00\x00.\x00"
    b"Daemon launched. Starting to output traces...\x00"
    b"DLT\x01U\xc4\x91Y\x8c>\n\x00MGHS5\x00\x00 MGHS"  # not to buffer
    b"\x00\x00mj&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
)

file_with_lifecycles_without_start = (
    b"DLT\x01\xc5\x82\xdaX\x19\x93\r\x00XORA'\x01\x00\x1bXORA"  # trace to buffer
    b"\x16\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11\x04\x00\x00\x00\x00"
    b"DLT\x01\xc5\x82\xdaXQi\x0e\x00MGHS5\x00\x00 MGHS"  # trace to buffer
    b"\x00\x03U\xe0&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
    b"DLT\x01m\xc2\x91Y\xad\xe4\x07\x00MGHS=\x01\x00zMGHS"  # random trace
    b"\x00\x00\x02\xab\x00\x00@VA\x01DLTDINTM\x00\x02\x00\x00Z\x00"
    b"ApplicationID 'DBSY' registered for PID 689, Description=DBus"
    b" Logging|SysInfra|Log&Trace\n\x00"
    b"DLT\x01\xed\xc2\x91Y\x0f\xf0\x08\x00MGHS5\x00\x00 MGHS"  # trace to buffer
    b"\x00\x00\x9dC&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
    b"DLT\x01\xed\xc2\x91Y\x17.\n\x00MGHS=\x00\x00NMGHS"  # new lifecycle
    b"\x00\x00\x02\xae\x00\x00@/A\x01DLTDINTM\x00\x02\x00\x00.\x00"
    b"Daemon launched. Starting to output traces...\x00"
)


def append_stream_to_file(stream, file_name):
    msgs = create_messages(stream, from_file=True)
    for msg in msgs:
        append_message_to_file(msg, file_name)


def append_message_to_file(message, file_name):
    # Use 'ab' instead of 'wb' because it is to append instead to overwrite
    with open(file_name, "ab") as file:
        file.write(message.to_bytes())
        file.flush()


def create_messages(stream, from_file=False):
    if from_file is False:
        stream.seek(0)
        buf = stream.read()

        client = DLTClient()
        client.receiver.buf = ctypes.create_string_buffer(buf)
        client.receiver.bytesRcvd = len(buf)

        return client.read_message()

    _, tmpname = tempfile.mkstemp(suffix=b"")
    tmpfile = open(tmpname, "wb")
    tmpfile.write(stream)
    tmpfile.flush()
    tmpfile.seek(0)
    tmpfile.close()

    atexit.register(os.remove, tmpname)

    msgs = load(tmpname)
    return msgs


class MockDLTMessage(object):
    """Mock DLT message for dltlyse plugin testing"""

    def __init__(self, ecuid="MGHS", apid="SYS", ctid="JOUR", sid="958", payload="", tmsp=0.0, sec=0, msec=0, mcnt=0):
        self.ecuid = ecuid
        self.apid = apid
        self.ctid = ctid
        self.sid = sid
        self.payload = payload
        self.tmsp = tmsp
        self.mcnt = mcnt
        self.storageheader = MockStorageHeader(sec=sec, msec=msec)

    def compare(self, target):
        """Compare DLT Message to a dictionary"""
        return target == {k: v for k, v in self.__dict__.items() if k in target.keys()}

    @property
    def payload_decoded(self):
        """Fake payload decoding"""
        return self.payload

    @property
    def storage_timestamp(self):
        """Fake storage timestamp"""
        return float("{}.{}".format(self.storageheader.seconds, self.storageheader.microseconds))

    def __repr__(self):
        return str(self.__dict__)


class MockStorageHeader(object):
    """Mock DLT storage header for plugin testing"""

    def __init__(self, msec=0, sec=0):
        self.microseconds = msec
        self.seconds = sec



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.

from dlt.helpers import LimitCondition


class TestsLimitCondition(object):
    __test__ = True

    def test_none(self):
        cond = LimitCondition(None)
        assert cond()

    def test_limit_decreasing(self):
        cond = LimitCondition(2)
        cond()
        assert cond.limit == 1
        assert cond()  # limit=0
        assert not cond()  # limit=-1



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.
"""Default implementation of the ctypes bindings for the DLT library"""

import ctypes
import logging
import sys
import ctypes.util

dlt_path = ctypes.util.find_library("dlt")

if sys.platform.startswith("darwin"):
    dltlib = ctypes.cdll.LoadLibrary(dlt_path or "libdlt.dylib")
elif sys.platform.startswith("linux"):
    if dlt_path:
        dltlib = ctypes.cdll.LoadLibrary(dlt_path)
    else:
        try:
            dltlib = ctypes.cdll.LoadLibrary("libdlt.so.3")
        except OSError:
            dltlib = ctypes.cdll.LoadLibrary("libdlt.so.2")
else:
    raise RuntimeError("Platform %s not supported" % sys.platform)

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Return value for DLTFilter.add() - exceeded maximum number of filters
MAX_FILTER_REACHED = 1
# Return value for DLTFilter.add() - specified filter already exists
REPEATED_FILTER = 2

DLT_ID_SIZE = 4
DLT_FILTER_MAX = 30  # Maximum number of filters
DLT_HTYP_UEH = 0x01  # use extended header
DLT_HTYP_WEID = 0x04  # with ECU ID
DLT_HTYP_WSID = 0x08  # with Session ID
DLT_HTYP_WTMS = 0x10  # with timestamp
DLT_MESSAGE_ERROR_OK = 0

DLT_DAEMON_TEXTSIZE = 10024

DLT_OUTPUT_HEX = 1
DLT_OUTPUT_ASCII = 2
DLT_OUTPUT_MIXED_FOR_PLAIN = 3
DLT_OUTPUT_MIXED_FOR_HTML = 4
DLT_OUTPUT_ASCII_LIMITED = 5

DLT_RETURN_ERROR = -1
DLT_RETURN_OK = 0
DLT_RETURN_TRUE = 1

# DltClientMode from dlt_client.h
DLT_CLIENT_MODE_UNDEFINED = -1
DLT_CLIENT_MODE_TCP = 0
DLT_CLIENT_MODE_SERIAL = 1
DLT_CLIENT_MODE_UNIX = 2
DLT_CLIENT_MODE_UDP_MULTICAST = 3

# DltReceiverType from dlt_common.h
# DltReceiverType is an enum type. These definitions could not be found in shared library (libdlt.so) so
# the enum values are defined here.
DLT_RECEIVE_SOCKET = 0
DLT_RECEIVE_FD = 1

DLT_TYPE_LOG = 0x00  # Log message type
DLT_TYPE_APP_TRACE = 0x01  # Application trace message type
DLT_TYPE_NW_TRACE = 0x02  # Network trace message type
DLT_TYPE_CONTROL = 0x03  # Control message type
DLT_CONTROL_REQUEST = 0x01
DLT_CONTROL_RESPONSE = 0x02  # Response to request message
DLT_CONTROL_TIME = 0x03
DLT_MSIN_MSTP_SHIFT = 1  # shift right offset to get mstp value
DLT_MSIN_MTIN_SHIFT = 4  # shift right offset to get mtin value
DLT_MSIN_MSTP = 0x0E  # message type
DLT_MSIN_MTIN = 0xF0  # message type info
DLT_MSIN_VERB = 0x01  # verbose mode
DLT_MSIN_CONTROL_RESPONSE = (DLT_TYPE_CONTROL << DLT_MSIN_MSTP_SHIFT) | (DLT_CONTROL_RESPONSE << DLT_MSIN_MTIN_SHIFT)

# dlt_protocol.h
DLT_SERVICE_ID_GET_SOFTWARE_VERSION = 0x13  # Service ID: Get software version
DLT_SERVICE_ID_UNREGISTER_CONTEXT = 0xF01  # Service ID: Message unregister context
DLT_SERVICE_ID_CONNECTION_INFO = 0xF02  # Service ID: Message connection info
DLT_SERVICE_ID_TIMEZONE = 0xF03  # Service ID: Timezone
DLT_SERVICE_ID_MARKER = 0xF04  # Service ID: Marker

DLT_CONNECTION_STATUS_DISCONNECTED = 0x01  # Client is disconnected
DLT_CONNECTION_STATUS_CONNECTED = 0x02  # Client is connected

DLT_TYPE_INFO_TYLE = 0x0000000F  # Length of standard data: 1 = 8bit, 2 = 16bit, 3 = 32 bit, 4 = 64 bit, 5 = 128 bit
DLT_TYPE_INFO_BOOL = 0x00000010  # Boolean data
DLT_TYPE_INFO_SINT = 0x00000020  # Signed integer data
DLT_TYPE_INFO_UINT = 0x00000040  # Unsigned integer data
DLT_TYPE_INFO_FLOA = 0x00000080  # Float data
DLT_TYPE_INFO_ARAY = 0x00000100  # Array of standard types
DLT_TYPE_INFO_STRG = 0x00000200  # String
DLT_TYPE_INFO_RAWD = 0x00000400  # Raw data
DLT_TYPE_INFO_VARI = 0x00000800  # Set, if additional information to a variable is available
DLT_TYPE_INFO_FIXP = 0x00001000  # Set, if quantization and offset are added
DLT_TYPE_INFO_TRAI = 0x00002000  # Set, if additional trace information is added
DLT_TYPE_INFO_STRU = 0x00004000  # Struct
DLT_TYPE_INFO_SCOD = 0x00038000  # coding of the type string: 0 = ASCII, 1 = UTF-8

DLT_SCOD_ASCII = 0x00000000
DLT_SCOD_UTF8 = 0x00008000
DLT_SCOD_HEX = 0x00010000
DLT_SCOD_BIN = 0x00018000

DLT_TYLE_8BIT = 0x00000001
DLT_TYLE_16BIT = 0x00000002
DLT_TYLE_32BIT = 0x00000003
DLT_TYLE_64BIT = 0x00000004
DLT_TYLE_128BIT = 0x00000005

DLT_DAEMON_TCP_PORT = 3490
DLT_CLIENT_RCVBUFSIZE = 10024  # Size of client receive buffer from dlt_client_cfg.h

# dlt-viever/qdltbase.cpp
qDltMessageType = [b"log", b"app_trace", b"nw_trace", b"control", b"", b"", b"", b""]
qDltLogInfo = [
    b"",
    b"fatal",
    b"error",
    b"warn",
    b"info",
    b"debug",
    b"verbose",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
]
qDltTraceType = [
    b"",
    b"variable",
    b"func_in",
    b"func_out",
    b"state",
    b"vfb",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
    b"",
]
qDltNwTraceType = [b"", b"ipc", b"can", b"flexray", b"most", b"vfb", b"", b"", b"", b"", b"", b"", b"", b"", b"", b""]
qDltControlType = [b"", b"request", b"response", b"time", b"", b"", b"", b"", b"", b"", b"", b"", b"", b"", b"", b""]
cqDltMode = [b"non-verbose", b"verbose"]
qDltEndianness = [b"little-endian", b"big-endian"]
cqDltTypeInfo = [
    b"String",
    b"Bool",
    b"SignedInteger",
    b"UnsignedInteger",
    b"Float",
    b"RawData",
    b"TraceInfo",
    b"Utf8String",
]
qDltCtrlServiceId = [
    b"",
    b"set_log_level",
    b"set_trace_status",
    b"get_log_info",
    b"get_default_log_level",
    b"store_config",
    b"reset_to_factory_default",
    b"set_com_interface_status",
    b"set_com_interface_max_bandwidth",
    b"set_verbose_mode",
    b"set_message_filtering",
    b"set_timing_packets",
    b"get_local_time",
    b"use_ecu_id",
    b"use_session_id",
    b"use_timestamp",
    b"use_extended_header",
    b"set_default_log_level",
    b"set_default_trace_status",
    b"get_software_version",
    b"message_buffer_overflow",
]
qDltCtrlReturnType = [b"ok", b"not_supported", b"error", b"3", b"4", b"5", b"6", b"7", b"no_matching_context_id"]


class cDltServiceConnectionInfo(ctypes.Structure):
    """
    typedef struct
    {
        uint32_t service_id;            /**< service ID */
        uint8_t status;                 /**< reponse status */
        uint8_t state;                  /**< new state */
        char comid[DLT_ID_SIZE];        /**< communication interface */
    } PACKED DltServiceConnectionInfo;
    """

    _fields_ = [
        ("service_id", ctypes.c_uint32),
        ("status", ctypes.c_uint8),
        ("state", ctypes.c_uint8),
        ("comid", DLT_ID_SIZE * ctypes.c_byte),
    ]
    _pack_ = 1
    _layout_ = "ms"


class MessageMode(object):
    """Default properties for the DLTMessage"""

    # pylint: disable=no-member

    @property
    def use_extended_header(self):
        """Returns True if the DLTMessage has extended header"""
        return self.standardheader.htyp & DLT_HTYP_UEH

    @property
    def _is_extended_header_exists(self):
        return self.extendedheader and self.extendedheader.msin

    @property
    def is_mode_verbose(self):
        """Returns True if the DLTMessage is set to verbose mode"""
        if not self._is_extended_header_exists:
            return self.verbose
        return self.extendedheader.msin & DLT_MSIN_VERB

    @property
    def mode_string(self):
        """Returns 'verbose' if DLTMessage is set to verbose mode. Otherwise 'non-verbose'"""
        return b"verbose" if self.is_mode_verbose else b"non-verbose"

    @property
    def is_mode_non_verbose(self):
        """Returns True if the DLTMessage is set to non-verbose mode"""
        return not self.is_mode_verbose

    @property
    def is_type_control(self):
        """Returns True if the DLTMessage type is control"""
        return self.standardheader.htyp & DLT_TYPE_CONTROL

    @property
    def is_type_control_response(self):
        """Returns True if the DLTMessage type is control response"""
        return self.standardheader.htyp & DLT_MSIN_CONTROL_RESPONSE

    @property
    def message_id(self):
        """Returns message ID of the DLTMessage"""
        if self.is_mode_non_verbose and (self.datasize >= 4):
            ptr_int = ctypes.cast(self.databuffer, ctypes.POINTER(ctypes.c_uint32))
            mid = ptr_int[0]
            return mid
        return 0

    @property
    def message_id_string(self):
        """Returns string representation of message ID"""
        mid = self.message_id
        return qDltCtrlServiceId[mid] if 0 <= mid <= len(qDltCtrlServiceId) else b""

    @property
    def ctrl_service_id(self):
        """Returns service ID of the DLTMessage"""
        service_id = 0
        if self.is_type_control and self.datasize >= 4:
            ptr_int = ctypes.cast(self.databuffer, ctypes.POINTER(ctypes.c_uint32))
            service_id = ptr_int[0]
        return service_id

    @property
    def ctrl_service_id_string(self):
        """Returns string representation of service ID"""
        sid = self.ctrl_service_id
        if sid == DLT_SERVICE_ID_UNREGISTER_CONTEXT:
            return b"unregister_context"
        if sid == DLT_SERVICE_ID_CONNECTION_INFO:
            return b"connection_info"
        if sid == DLT_SERVICE_ID_TIMEZONE:
            return b"timezone"
        if sid == DLT_SERVICE_ID_MARKER:
            return b"marker"
        return qDltCtrlServiceId[sid] if sid <= 20 else b""

    @property
    def ctrl_return_type(self):
        """Returns ctrl type of the DLTMessage"""
        return_type = 0
        if self.is_type_control and (self.is_type_control_response and self.datasize >= 6):
            return_type = self.databuffer[4]
        return return_type

    @property
    def ctrl_return_type_string(self):
        """Returns string representation of ctrl type"""
        return qDltCtrlReturnType[self.ctrl_return_type] if self.ctrl_return_type <= 8 else b""

    @property
    def type(self):
        """Returns message type of the DLTMessage"""
        if not self._is_extended_header_exists:
            return DLT_TYPE_LOG
        return (self.extendedheader.msin & DLT_MSIN_MSTP) >> DLT_MSIN_MSTP_SHIFT

    @property
    def type_string(self):
        """Returns string representation of the message type"""
        mtype = self.type
        return qDltMessageType[mtype] if 0 <= mtype <= 7 else b""

    @property
    def subtype(self):
        """Returns message subtype of the DLTMessage"""
        if not self._is_extended_header_exists:
            return DLT_TYPE_LOG
        return (self.extendedheader.msin & DLT_MSIN_MTIN) >> DLT_MSIN_MTIN_SHIFT

    @property
    def subtype_string(self):
        """Returns string representation of the message subtype"""
        mtype = self.type
        msubtype = self.subtype

        if mtype == DLT_TYPE_LOG:
            return qDltLogInfo[msubtype] if 0 <= msubtype <= 7 else b""
        if mtype == DLT_TYPE_APP_TRACE:
            return qDltTraceType[msubtype] if 0 <= msubtype <= 7 else b""
        if mtype == DLT_TYPE_NW_TRACE:
            return qDltNwTraceType[msubtype] if 0 <= msubtype <= 7 else b""
        if mtype == DLT_TYPE_CONTROL:
            return qDltControlType[msubtype] if 0 <= msubtype <= 7 else b""

        return b""

    @property
    def payload_decoded(self):
        """Decode the payload data

        :returns: Payload data
        :rtype: str
        """
        text = b""
        if self.is_mode_non_verbose and not self.is_type_control and self.noar == 0:
            buf = ctypes.create_string_buffer(b"\000" * DLT_DAEMON_TEXTSIZE)
            dltlib.dlt_message_payload(ctypes.byref(self), buf, DLT_DAEMON_TEXTSIZE, DLT_OUTPUT_ASCII, self.verbose)
            return b"[%s] #%s#" % (self.message_id_string, buf.value[4:].rstrip(b"\000"))

        if self.type == DLT_TYPE_CONTROL and self.subtype == DLT_CONTROL_RESPONSE:
            if self.ctrl_service_id == DLT_SERVICE_ID_MARKER:
                return b"MARKER"

            text = b"[%s %s] " % (self.ctrl_service_id_string, self.ctrl_return_type_string)
            service_id = self.ctrl_service_id

            if self.ctrl_service_id == DLT_SERVICE_ID_GET_SOFTWARE_VERSION:
                text += ctypes.string_at(self.databuffer, self.datasize)[9:].rstrip(b"\000")
            elif self.ctrl_service_id == DLT_SERVICE_ID_CONNECTION_INFO:
                if self.datasize == ctypes.sizeof(cDltServiceConnectionInfo):
                    conn_info = cDltServiceConnectionInfo.from_buffer(bytearray(self.databuffer[: self.datasize]))
                    if conn_info.state == DLT_CONNECTION_STATUS_DISCONNECTED:
                        text += b"disconnected"
                    elif conn_info.state == DLT_CONNECTION_STATUS_CONNECTED:
                        text += b"connected"
                    else:
                        text += b"unknown"
                    text += b" " + ctypes.string_at(conn_info.comid, DLT_ID_SIZE).rstrip(b"\000")
                else:
                    text += ctypes.string_at(self.databuffer, self.datasize)[5 : 256 + 5].rstrip(b"\000")
            elif service_id == DLT_SERVICE_ID_TIMEZONE:
                text += ctypes.string_at(self.databuffer, self.datasize)[5 : 256 + 5].rstrip(b"\000")
            else:
                buf = ctypes.create_string_buffer(b"\000" * DLT_DAEMON_TEXTSIZE)
                dltlib.dlt_message_payload(
                    ctypes.byref(self), buf, DLT_DAEMON_TEXTSIZE, DLT_OUTPUT_ASCII, self.verbose
                )
                text += buf.value.rstrip(b"\000")
            return text

        if self.type == DLT_TYPE_CONTROL:
            return b"[%s] %s" % (
                self.ctrl_service_id_string,
                ctypes.string_at(self.databuffer, self.datasize)[4 : 256 + 4].rstrip(b"\000"),
            )

        buf = ctypes.create_string_buffer(b"\000" * DLT_DAEMON_TEXTSIZE)
        dltlib.dlt_message_payload(ctypes.byref(self), buf, DLT_DAEMON_TEXTSIZE, DLT_OUTPUT_ASCII, self.verbose)
        return buf.value.rstrip(b"\000").strip()


class cDltStorageHeader(ctypes.Structure):
    """
    /**
     * The structure of the DLT file storage header. This header is used before each stored DLT message.
     */
    typedef struct
    {
        char pattern[DLT_ID_SIZE];        /**< This pattern should be DLT0x01 */
        uint32_t seconds;                    /**< seconds since 1.1.1970 */
        int32_t microseconds;            /**< Microseconds */
        char ecu[DLT_ID_SIZE];            /**< The ECU id is added, if it is not already in the DLT message itself */
    } PACKED DltStorageHeader;
    """

    _fields_ = [
        ("pattern", ctypes.c_char * DLT_ID_SIZE),
        ("seconds", ctypes.c_uint32),
        ("microseconds", ctypes.c_int32),
        ("ecu", ctypes.c_char * DLT_ID_SIZE),
    ]
    _pack_ = 1
    _layout_ = "ms"

    def __reduce__(self):
        return (cDltStorageHeader, (self.pattern, self.seconds, self.microseconds, self.ecu))


class cDltStandardHeader(ctypes.BigEndianStructure):
    """The structure of the DLT standard header. This header is used in each DLT message.

    typedef struct
    {
        uint8_t htyp;           /**< This parameter contains several informations, see definitions below */
        uint8_t mcnt;           /**< The message counter is increased with each sent DLT message */
        uint16_t len;           /**< Length of the complete message, without storage header */
    } PACKED DltStandardHeader;
    """

    _fields_ = [("htyp", ctypes.c_uint8), ("mcnt", ctypes.c_uint8), ("len", ctypes.c_ushort)]
    _pack_ = 1
    _layout_ = "ms"

    def __reduce__(self):
        return (cDltStandardHeader, (self.htyp, self.mcnt, self.len))


class cDltStandardHeaderExtra(ctypes.Structure):
    """The structure of the DLT extra header parameters. Each parameter is sent only if enabled in htyp.

    typedef struct
    {
        char ecu[DLT_ID_SIZE];       /**< ECU id */
        uint32_t seid;     /**< Session number */
        uint32_t tmsp;     /**< Timestamp since system start in 0.1 milliseconds */
    } PACKED DltStandardHeaderExtra;
    """

    _fields_ = [("ecu", ctypes.c_char * DLT_ID_SIZE), ("seid", ctypes.c_uint32), ("tmsp", ctypes.c_uint32)]
    _pack_ = 1
    _layout_ = "ms"

    def __reduce__(self):
        return (cDltStandardHeaderExtra, (self.ecu, self.seid, self.tmsp))


class cDltExtendedHeader(ctypes.Structure):
    """The structure of the DLT extended header. This header is only sent if enabled in htyp parameter.

    typedef struct
    {
        uint8_t msin;          /**< messsage info */
        uint8_t noar;          /**< number of arguments */
        char apid[DLT_ID_SIZE];          /**< application id */
        char ctid[DLT_ID_SIZE];          /**< context id */
    } PACKED DltExtendedHeader;
    """

    _fields_ = [
        ("msin", ctypes.c_uint8),
        ("noar", ctypes.c_uint8),
        ("apid", ctypes.c_char * DLT_ID_SIZE),
        ("ctid", ctypes.c_char * DLT_ID_SIZE),
    ]
    _pack_ = 1
    _layout_ = "ms"

    def __reduce__(self):
        return (cDltExtendedHeader, (self.msin, self.noar, self.apid, self.ctid))


class cDLTMessage(ctypes.Structure):
    """The structure of the DLT messages.

    typedef struct sDltMessage
    {
        /* flags */
        int8_t found_serialheader;

        /* offsets */
        int32_t resync_offset;

        /* size parameters */
        int32_t headersize;    /**< size of complete header including storage header */
        int32_t datasize;      /**< size of complete payload */

        /* buffer for current loaded message */
        uint8_t headerbuffer[sizeof(DltStorageHeader)+
                             sizeof(DltStandardHeader)+sizeof(DltStandardHeaderExtra)+sizeof(DltExtendedHeader)];
                             /**< buffer for loading complete header */
        uint8_t *databuffer;         /**< buffer for loading payload */
        int32_t databuffersize;

        /* header values of current loaded message */
        DltStorageHeader       *storageheader;  /**< pointer to storage header of current loaded header */
        DltStandardHeader      *standardheader; /**< pointer to standard header of current loaded header */
        DltStandardHeaderExtra headerextra;     /**< extra parameters of current loaded header */
        DltExtendedHeader      *extendedheader; /**< pointer to extended of current loaded header */
    } DltMessage;
    """

    _fields_ = [
        ("found_serialheader", ctypes.c_int8),
        ("resync_offset", ctypes.c_int32),
        ("headersize", ctypes.c_int32),
        ("datasize", ctypes.c_int32),
        (
            "headerbuffer",
            ctypes.c_uint8
            * (
                ctypes.sizeof(cDltStorageHeader)
                + ctypes.sizeof(cDltStandardHeader)
                + ctypes.sizeof(cDltStandardHeaderExtra)
                + ctypes.sizeof(cDltExtendedHeader)
            ),
        ),
        ("databuffer", ctypes.POINTER(ctypes.c_uint8)),
        ("databuffersize", ctypes.c_uint32),
        ("p_storageheader", ctypes.POINTER(cDltStorageHeader)),
        ("p_standardheader", ctypes.POINTER(cDltStandardHeader)),
        ("headerextra", cDltStandardHeaderExtra),
        ("p_extendedheader", ctypes.POINTER(cDltExtendedHeader)),
    ]


class cDltReceiver(ctypes.Structure):
    """The structure is used to organise the receiving of data including buffer handling.
    This structure is used by the corresponding functions.

    typedef struct
    {
        int32_t lastBytesRcvd;    /**< bytes received in last receive call */
        int32_t bytesRcvd;        /**< received bytes */
        int32_t totalBytesRcvd;   /**< total number of received bytes */
        char *buffer;         /**< pointer to receiver buffer */
        char *buf;            /**< pointer to position within receiver buffer */
        int fd;               /**< connection handle */
        int32_t buffersize;       /**< size of receiver buffer */
    } DltReceiver;
    """

    _fields_ = [
        ("lastBytesRcvd", ctypes.c_int32),
        ("bytesRcvd", ctypes.c_int32),
        ("totalBytesRcvd", ctypes.c_int32),
        ("buffer", ctypes.POINTER(ctypes.c_char)),
        ("buf", ctypes.POINTER(ctypes.c_char)),
        ("fd", ctypes.c_int),
        ("buffersize", ctypes.c_int32),
    ]


class cDltClient(ctypes.Structure):
    """
    typedef struct
    {
        DltReceiver receiver;  /**< receiver pointer to dlt receiver structure */
        int sock;              /**< sock Connection handle/socket */
        char *servIP;          /**< servIP IP adress/Hostname of TCP/IP interface */
        char *serialDevice;    /**< serialDevice Devicename of serial device */
        speed_t baudrate;      /**< baudrate Baudrate of serial interface, as speed_t */
        int serial_mode;       /**< serial_mode Serial mode enabled =1, disabled =0 */
    } DltClient;
    """

    _fields_ = [
        ("receiver", cDltReceiver),
        ("sock", ctypes.c_int),
        ("servIP", ctypes.c_char_p),
        ("serialDevice", ctypes.c_char_p),
        ("baudrate", ctypes.c_int),
        ("serial_mode", ctypes.c_int),
    ]


class cDLTFilter(ctypes.Structure):  # pylint: disable=invalid-name
    """
    typedef struct
    {
        char apid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< application id */
        char ctid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< context id */
        int  counter;                           /**< number of filters */
    } DltFilter;
    """

    _fields_ = [
        ("apid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
        ("ctid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
        ("counter", ctypes.c_int),
    ]

    # pylint: disable=too-many-arguments
    def add(self, apid, ctid):
        """Add new filter pair"""
        if isinstance(apid, str):
            apid = bytes(apid, "ascii")
        if isinstance(ctid, str):
            ctid = bytes(ctid, "ascii")
        if dltlib.dlt_filter_add(ctypes.byref(self), apid or b"", ctid or b"", self.verbose) == DLT_RETURN_ERROR:
            if self.counter >= DLT_FILTER_MAX:
                logger.error("Maximum number (%d) of allowed filters reached, ignoring filter!\n", DLT_FILTER_MAX)
                return MAX_FILTER_REACHED
            logger.debug("Filter ('%s', '%s') already exists", apid, ctid)
            return REPEATED_FILTER
        return 0



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.
import os
import time
import unittest
from queue import Empty
from multiprocessing import Event, Queue

from dlt.dlt_broker_handlers import DLTMessageHandler
from tests.utils import create_messages, stream_multiple


class TestDLTMessageHandler(unittest.TestCase):
    def setUp(self):
        self.filter_queue = Queue()
        self.message_queue = Queue()
        self.client_cfg = {
            "ip_address": b"127.0.0.1",
            "filename": b"/dev/null",
            "verbose": 0,
            "port": "1234",
        }
        self.stop_event = Event()
        self.handler = DLTMessageHandler(self.filter_queue, self.message_queue, self.stop_event, self.client_cfg)

    def test_init(self):
        self.assertFalse(self.handler.mp_stop_flag.is_set())
        self.assertFalse(self.handler.is_alive())
        self.assertTrue(self.handler.filter_queue.empty())
        self.assertTrue(self.handler.message_queue.empty())

    def test_run_basic(self):
        self.assertFalse(self.handler.is_alive())
        self.handler.start()
        self.assertTrue(self.handler.is_alive())
        self.assertNotEqual(self.handler.pid, os.getpid())
        self.stop_event.set()
        self.handler.join()
        self.assertFalse(self.handler.is_alive())

    def test_handle_add_new_filter(self):
        self.handler.filter_queue.put(("queue_id", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.handler.handle(None)
        self.assertIn(("SYS", "JOUR"), self.handler.context_map)
        self.assertEqual(self.handler.context_map[("SYS", "JOUR")], ["queue_id"])

    def test_handle_remove_filter_single_entry(self):
        self.handler.filter_queue.put(("queue_id", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.handler.handle(None)
        self.assertIn(("SYS", "JOUR"), self.handler.context_map)
        self.assertEqual(self.handler.context_map[("SYS", "JOUR")], ["queue_id"])

        self.handler.filter_queue.put(("queue_id", [("SYS", "JOUR")], False))
        time.sleep(0.01)
        self.handler.handle(None)
        self.assertNotIn(("SYS", "JOUR"), self.handler.context_map)

    def test_handle_remove_filter_multiple_entries(self):
        self.handler.filter_queue.put(("queue_id1", [("SYS", "JOUR")], True))
        self.handler.filter_queue.put(("queue_id2", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.handler.handle(None)
        self.assertIn(("SYS", "JOUR"), self.handler.context_map)
        self.assertEqual(self.handler.context_map[("SYS", "JOUR")], ["queue_id1", "queue_id2"])

        self.handler.filter_queue.put(("queue_id1", [("SYS", "JOUR")], False))
        time.sleep(0.01)
        self.handler.handle(None)
        self.assertIn(("SYS", "JOUR"), self.handler.context_map)
        self.assertEqual(self.handler.context_map[("SYS", "JOUR")], ["queue_id2"])

    def test_handle_multiple_similar_filters(self):
        self.handler.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.handler.filter_queue.put(("queue_id1", [("SYS", "JOUR")], True))
        time.sleep(0.01)
        self.handler.handle(None)
        self.assertIn(("SYS", "JOUR"), self.handler.context_map)
        self.assertEqual(self.handler.context_map[("SYS", "JOUR")], ["queue_id0", "queue_id1"])

    def test_handle_multiple_different_filters(self):
        self.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.filter_queue.put(("queue_id1", [("DA1", "DC1")], True))
        time.sleep(0.01)
        self.handler.handle(None)
        self.assertIn(("SYS", "JOUR"), self.handler.context_map)
        self.assertIn(("DA1", "DC1"), self.handler.context_map)
        self.assertEqual(self.handler.context_map[("SYS", "JOUR")], ["queue_id0"])
        self.assertEqual(self.handler.context_map[("DA1", "DC1")], ["queue_id1"])

    def test_handle_message_tag_and_distribute(self):
        self.filter_queue.put(("queue_id0", [("SYS", "JOUR")], True))
        self.filter_queue.put(("queue_id1", [("DA1", "DC1")], True))
        self.filter_queue.put(("queue_id2", [("SYS", None)], True))
        self.filter_queue.put(("queue_id3", [(None, "DC1")], True))
        self.filter_queue.put(("queue_id4", [(None, None)], True))
        time.sleep(0.01)

        # - simulate receiving of messages
        for _ in range(10):
            for message in create_messages(stream_multiple, from_file=True):
                self.handler.handle(message)

        self.assertIn(("SYS", "JOUR"), self.handler.context_map)
        self.assertIn(("DA1", "DC1"), self.handler.context_map)
        self.assertIn((None, None), self.handler.context_map)
        self.assertIn(("SYS", None), self.handler.context_map)
        self.assertIn((None, "DC1"), self.handler.context_map)
        try:
            # 60 == 10 messages of each for SYS, JOUR and None combinations +
            #       10 for (None,None)
            messages = [self.message_queue.get(timeout=0.01) for _ in range(60)]

            # these queues should not get any messages from other queues
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id0"]), 10)
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id1"]), 10)
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id2"]), 10)
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id3"]), 10)
            # this queue should get all messages
            self.assertEqual(len([msg for qid, msg in messages if qid == "queue_id4"]), 20)
        except Empty:
            # - we should not get an Empty for at least 40 messages
            self.fail()



# Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
"""DLT client helpers"""


class LimitCondition(object):
    """Condition object for counting messages"""

    def __init__(self, limit):
        """Constructor

        :param int limit: The maximum number of the messages for the condition
        """
        self.limit = limit

    def __call__(self):
        if self.limit is None:
            return True

        self.limit = self.limit - 1
        return self.limit >= 0


class ContinuousnessChecker(object):
    """ContinuousnessChecker class is intended to find problems in the order of DLT messages"""

    _ignore = ["DA1-DC1-0"]  # control message will be ignored - there is no continuation

    def __init__(self, start=0):
        self._index = start
        self._counter = dict()

    def __call__(self, message):
        key = "{}-{}-{}".format(message.apid, message.ctid, message.seid)

        self._index += 1

        if key in self._ignore:
            return

        if key in self._counter:
            # message of current type already received - check the continuousness
            err_msg = "Missing message detected. Message"
            err_msg += " #{} (apid='%s', ctid='%s', seid='%s')" % (message.apid, message.ctid, message.seid)
            err_msg += " should have counter '{}' instead of '{}'"

            if not (self._counter[key] + 1) % 256 == message.mcnt:
                counter = self._counter[key]
                self._counter[key] = message.mcnt
                raise RuntimeError(err_msg.format(self._index - 1, (counter + 1) % 256, message.mcnt))

            self._counter[key] = message.mcnt
        else:
            # first message of current type
            self._counter[key] = message.mcnt


def bytes_to_str(byte_or_str):
    """Return string from bytes"""
    if isinstance(byte_or_str, bytes):
        return byte_or_str.decode("utf8", errors="replace")

    return str(byte_or_str)



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.
"""Basic unittests for DLT messages"""

import io
import unittest

from dlt.dlt import DLTFilter

from .utils import create_messages

stream_one = io.BytesIO(b"5\x00\x00 MGHS\xdd\xf6e\xca&\x01DA1\x00DC1\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00")
stream_two = io.BytesIO(b"5\x00\x00 MGHS\xdd\xf6e\xca&\x01DA1\x00DC2\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00")

LOOPS = 100000


class TestsDLTMessagePerf(unittest.TestCase):
    def setUp(self):
        self.msgs = [create_messages(stream_one) for i in range(int(LOOPS * 0.1))]
        self.msgs += [create_messages(stream_two) for i in range(int(LOOPS * 0.9))]

    def test_compare_dict(self):
        # with dict as other
        attrs = {"apid": "DA1", "ctid": "DC1"}
        for msg in self.msgs:
            msg.compare(other=attrs)

    def test_compare_filter(self):
        # with DLTFilter as other
        flt = DLTFilter()
        flt.add("DA1", "DC1")
        for msg in self.msgs:
            msg.compare(other=flt)

    def test_compare_mesage(self):
        # with dict as other
        other = create_messages(stream_one)
        for msg in self.msgs:
            msg.compare(other=other)



# Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
"""Basic unittests for DLT messages"""

import io
import pickle
import re
from unittest.mock import patch, PropertyMock

import pytest

from dlt.dlt import DLTMessage
from tests.utils import (
    create_messages,
    stream_one,
    stream_with_params,
    stream_multiple,
    stream_multiple_with_malformed_message_at_begining,
    msg_benoit,
    control_one,
)


class TestsDLTMessageUnit(object):
    def test_malformed_message(self):
        msgs = create_messages(stream_multiple_with_malformed_message_at_begining, from_file=True)

        assert msgs[0].message_id == 1279675715
        assert len(msgs) == 3
        assert not msgs[0].extendedheader

    def test_compare_default_attrs(self):
        attrs = {"extendedheader.apid": "DA1", "extendedheader.ctid": "DC1"}
        msg = create_messages(stream_one)

        assert msg.compare(other=attrs)
        assert msg.compare(other={"extendedheader.ctid": "DC1"})

    def test_equal(self):
        msg1 = create_messages(stream_one)
        msg2 = create_messages(stream_one)

        assert msg1 == msg2

    def test_easy_attributes(self):
        msg = create_messages(stream_one)

        assert msg.ecuid == "MGHS"
        assert msg.seid == 0
        assert msg.tmsp == 372391.26500000001
        assert msg.apid == "DA1"
        assert msg.ctid == "DC1"

    def test_compare(self):
        msg1 = create_messages(stream_one)
        msg2 = create_messages(stream_one)

        assert msg1.compare(msg2)
        assert msg1.compare(other=msg2)
        assert msg1.compare(dict(apid="DA1", ctid="DC1"))
        assert not msg1.compare(dict(apid="DA1", ctid="XX"))

    def test_compare_regexp(self):
        msg1 = create_messages(stream_one)

        assert msg1.compare(dict(apid="DA1", ctid=re.compile(r"D.*")))
        assert msg1.compare(
            dict(apid="DA1", ctid=re.compile(r"D.*"), payload_decoded=re.compile(r".connection_info ok."))
        )
        assert msg1.compare(
            dict(apid="DA1", ctid=re.compile(r"D.*"), payload_decoded=re.compile(r".connection_info ok."))
        )
        assert msg1.compare(dict(apid="DA1", ctid=re.compile(r"D.*"), payload_decoded=re.compile(r".*info ok.")))
        assert msg1.compare(dict(apid="DA1", ctid="DC1", payload_decoded=re.compile(r".*info ok.")))
        assert msg1.compare(dict(apid=re.compile(r"D.")))
        assert msg1.compare(dict(apid=re.compile(r"D.+")))
        assert msg1.compare(dict(apid=re.compile(r"D.")))
        assert not msg1.compare(dict(apid=re.compile(r"X.")))

    def test_compare_regexp_nsm(self):
        nsm = create_messages(
            io.BytesIO(b"5\x00\x00 MGHS\xdd\xf6e\xca&\x01NSM\x00DC1\x00\x02\x0f\x00\x00" b"\x00\x02\x00\x00\x00\x00")
        )
        nsma = create_messages(
            io.BytesIO(b"5\x00\x00 MGHS\xdd\xf6e\xca&\x01NSMADC1\x00\x02\x0f\x00\x00" b"\x00\x02\x00\x00\x00\x00")
        )

        assert nsm.compare(dict(apid=re.compile("^NSM$")))
        assert not nsma.compare(dict(apid=re.compile("^NSM$")))

        assert nsm.compare(dict(apid="NSM"))
        assert not nsma.compare(dict(apid="NSM"))

        assert nsm.compare(dict(apid=re.compile("NSM")))
        assert nsma.compare(dict(apid=re.compile("NSM")))

    def test_compare_regexp_throw(self):
        nsm = create_messages(
            io.BytesIO(b"5\x00\x00 MGHS\xdd\xf6e\xca&\x01NSM\x00DC1\x00\x02\x0f\x00\x00" b"\x00\x02\x00\x00\x00\x00")
        )
        with pytest.raises(Exception):
            assert nsm.compare(dict(apid=b"NSM"), regexp=True)

    def test_compare_regexp_benoit(self):
        msg1 = create_messages(msg_benoit, from_file=True)[0]
        assert msg1.compare(
            {
                "apid": "DEMO",
                "ctid": "DATA",
                "payload_decoded": re.compile("Logging from the constructor of a global instance"),
            }
        )

    def test_compare_two_msgs(self):
        msgs = create_messages(stream_multiple, from_file=True)
        assert msgs[0] != msgs[-1]

    def test_compare_other_not_modified(self):
        msg = create_messages(stream_one)
        other = dict(apid="XX", ctid="DC1")
        assert not msg.compare(other)
        assert other == dict(apid="XX", ctid="DC1")

    def test_compare_quick_return(self):
        msg = create_messages(stream_one)
        other = dict(apid=b"DA1", ctid=b"XX", ecuid=b"FOO")

        with patch("dlt.dlt.DLTMessage.ecuid", new_callable=PropertyMock) as ecuid:
            ecuid.return_value = b"FOO"
            assert not msg.compare(other)
            ecuid.assert_not_called()

    def test_compare_matching_apid_ctid(self):
        msg = create_messages(stream_one)
        other = dict(apid="DA1", ctid="DC1", ecuid="FOO")

        with patch("dlt.dlt.DLTMessage.ecuid", new_callable=PropertyMock) as ecuid:
            ecuid.return_value = "BAR"
            assert not msg.compare(other)
            ecuid.assert_called_once()

            ecuid.return_value = "FOO"
            assert msg.compare(other)
            assert ecuid.call_count == 2

    def test_pickle_api(self):
        messages = create_messages(stream_multiple, from_file=True)
        for msg in messages:
            assert msg == pickle.loads(pickle.dumps(msg))

    def test_from_bytes_control(self):
        msg = DLTMessage.from_bytes(
            b"DLT\x011\xd9PY(<\x08\x00MGHS5\x00\x00 MGHS\x00\x00\x96\x85&\x01DA1\x00DC1"
            b"\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
        )

        assert msg.apid == "DA1"
        assert msg.ctid == "DC1"
        assert msg.ecuid == "MGHS"
        assert msg.tmsp == 3.8533
        assert msg.storage_timestamp == 1498470705.539688
        assert msg.payload_decoded == "[connection_info ok] connected "

    def test_from_bytes_log_multipayload(self):
        msg = DLTMessage.from_bytes(
            b"DLT\x011\xd9PYfI\x08\x00MGHS=\x00\x000MGHS\x00\x00\x03\x1e\x00\x00\x94\xc8A"
            b"\x01MON\x00CPUS\x00\x02\x00\x00\x10\x004 online cores\n\x00"
        )

        assert msg.apid == "MON"
        assert msg.ctid == "CPUS"
        assert msg.ecuid == "MGHS"
        assert msg.tmsp == 3.8088
        assert msg.payload_decoded == "4 online cores"

    def test_sort_data_control(self):
        data = (
            b"DLT\x011\xd9PY(<\x08\x00MGHS5\x00\x00 MGHS\x00\x00\x96\x85&\x01DA1\x00DC1"
            b"\x00\x02\x0f\x00\x00\x00\x02\x00\x00\x00\x00"
        )
        tmsp, length, apid, ctid = DLTMessage.extract_sort_data(data)

        assert tmsp == 3.8533
        assert length == len(data)
        assert apid == "DA1"
        assert ctid == "DC1"

    def test_sort_data_log_multipayload(self):
        data = (
            b"DLT\x011\xd9PYfI\x08\x00MGHS=\x00\x000MGHS\x00\x00\x03\x1e\x00\x00\x94\xc8A"
            b"\x01MON\x00CPUS\x00\x02\x00\x00\x10\x004 online cores\n\x00"
        )
        tmsp, length, apid, ctid = DLTMessage.extract_sort_data(data)

        assert tmsp == 3.8088
        assert length == len(data)
        assert apid == "MON"
        assert ctid == "CPUS"

    def test_largelog(self):
        data = (
            b"DLT\x012\xd9PY)\x00\x01\x00MGHS=o\x02\x04MGHS\x00\x00\x03\x1e\x00\x00\x9e\xb7"
            b"A\x01MON\x00THRD\x00\x02\x00\x00\xe4\x01Process avb_streamhandl with pid: 307 "
            b'"/usr/bin/avb_streamhandler_app_someip -s pluginias-media_transport-avb_config'
            b"uration_bmw_mgu.so --bg setup --target Harman_MGU_B1 -p MGU_ICAM -k local.alsa"
            b".baseperiod=256 -k ptp.loopcount=0 -k ptp.pdelaycount=0 -k ptp.synccount=0 -k "
            b"sched.priority=20 -k tspec.vlanprio.low=3 -k tspec.presentation.time.offset.lo"
            b"w=2200000 -k tspec.interval.low=1333000 -k debug.loglevel._RXE=4 -k alsa.group"
            b'name=mgu_avbsh -n socnet0 -b 2 "  started 2401 msec ago\x00'
        )

        msg = DLTMessage.from_bytes(data)
        assert msg.apid == "MON"
        assert msg.ctid == "THRD"
        assert msg.ecuid == "MGHS"
        assert msg.tmsp == 4.0631
        assert (
            msg.payload_decoded == 'Process avb_streamhandl with pid: 307 "/usr/bin/avb_streamhandler_app_someip -s '
            "pluginias-media_transport-avb_configuration_bmw_mgu.so --bg setup --target Harman_MGU_B1 -p MGU_ICAM "
            "-k local.alsa.baseperiod=256 -k ptp.loopcount=0 -k ptp.pdelaycount=0 -k ptp.synccount=0 "
            "-k sched.priority=20 -k tspec.vlanprio.low=3 -k tspec.presentation.time.offset.low=2200000 "
            "-k tspec.interval.low=1333000 -k debug.loglevel._RXE=4 -k alsa.groupname=mgu_avbsh -n socnet0 "
            '-b 2 "  started 2401 msec ago'
        )

        tmsp, length, apid, ctid = DLTMessage.extract_sort_data(data)
        assert msg.tmsp == tmsp
        assert len(msg.to_bytes()) == length
        assert msg.apid == apid
        assert msg.ctid == ctid


class TestsPayload(object):
    def test_split(self):
        msg = create_messages(stream_with_params, from_file=True)[0]
        payload = msg.payload
        assert len(payload) == msg.noar
        assert payload[0] == b"CLevelMonitor::notification() => commandType"
        assert payload[1] == 3
        assert payload[2] == b"deviceId"
        assert payload[3] == 5
        assert payload[4] == b"value"
        assert payload[5] == 4074
        assert payload[6] == b"simulation status"
        assert payload[7] == 0

        with pytest.raises(IndexError):
            payload.__getitem__(8)


class TestsControl(object):
    def test_load(self):
        msg = create_messages(control_one, from_file=True)[0]
        assert msg.apid == "DA1"
        assert msg.ctid == "DC1"
        assert msg.is_mode_verbose == 0
        assert (
            msg.payload_decoded == "[get_log_info 7] get_log_info, 07, 01 00 48 44 44 4d 01 00 43 41 50 49 ff"
            " ff 04 00 43 41 50 49 06 00 68 64 64 6d 67 72 72 65 6d 6f"
        )



# Copyright (C) 2016. BMW Car IT GmbH. All rights reserved.
"""Basic unittests for DLTClient class"""

import unittest
from unittest.mock import patch, Mock

from dlt.dlt import DLTClient, DLT_RETURN_OK, DLT_RETURN_ERROR


class TestDLTClient(unittest.TestCase):
    def setUp(self):
        # - patch port so that connect fails even if dlt-daemon is running
        self.client = DLTClient(servIP="127.0.0.1", port=424242)

    def test_connect_with_timeout_failed(self):
        # - timeout error
        self.assertFalse(self.client.connect(timeout=2))

        # - dlt_receiver_init error
        with (
            patch("socket.create_connection", return_value=Mock(fileno=Mock(return_value=2000000))),
            patch("dlt.dlt.dltlib.dlt_receiver_init", return_value=DLT_RETURN_ERROR),
        ):
            self.assertFalse(self.client.connect(timeout=2))

    def test_connect_with_timeout_success(self):
        with (
            patch("socket.create_connection", return_value=Mock(fileno=Mock(return_value=2000000))),
            patch("dlt.dlt.dltlib.dlt_receiver_init", return_value=DLT_RETURN_OK),
        ):
            self.assertTrue(self.client.connect(timeout=2))



from dlt.helpers import ContinuousnessChecker
import pytest


class Msg(object):
    def __init__(self, apid, ctid, seid, mcnt):
        self.apid = apid
        self.ctid = ctid
        self.seid = seid
        self.mcnt = mcnt


def run_check(messages):
    cont = ContinuousnessChecker()
    for msg in messages:
        cont(msg)


class TestsContinuousness(object):
    def test_simple(self):
        messages = [
            Msg("X", "Y", "99", 4),
            Msg("X", "Y", "99", 5),
            Msg("X", "Y", "99", 6),
            Msg("X", "Y", "99", 7),
            Msg("X", "Y", "99", 8),
        ]
        run_check(messages)

    def test_simple_missing(self):
        with pytest.raises(RuntimeError):
            messages = [
                Msg("X", "Y", "99", 4),
                Msg("X", "Y", "99", 5),
                Msg("X", "Y", "99", 6),
                # 7 is missing
                Msg("X", "Y", "99", 8),
                Msg("X", "Y", "99", 9),
            ]
            run_check(messages)

    def test_simple_over(self):
        # message counter is a unsigned char so counts till 255 and then restarted back to 0
        messages = [Msg("X", "Y", "99", 254), Msg("X", "Y", "99", 255), Msg("X", "Y", "99", 0), Msg("X", "Y", "99", 1)]
        run_check(messages)

    def test_simple_reset(self):
        with pytest.raises(RuntimeError):
            messages = [Msg("X", "Y", "99", 230), Msg("X", "Y", "99", 231), Msg("X", "Y", "99", 0)]
            run_check(messages)

    def test_ignore_control(self):
        messages = [Msg("DA1", "DC1", "0", 0), Msg("X", "Y", "99", 231), Msg("DA1", "DC1", "0", 0)]
        run_check(messages)

    def test_zeros_da1_dc1(self):
        messages = [Msg("DA1", "DC1", "0", 0), Msg("DA1", "DC1", "0", 0)]
        run_check(messages)

    def test_zeros_non_da1_dc1(self):
        with pytest.raises(RuntimeError):
            messages = [Msg("X", "Y", "0", 0), Msg("X", "Y", "0", 0)]
            run_check(messages)



# Copyright (C) 2023. BMW Car IT GmbH. All rights reserved.
"""Basic unittests for the py_dlt_file_main_loop function"""

import os
import unittest
import tempfile
from threading import Thread
import time

from dlt.dlt import cDLTFile, py_dlt_file_main_loop
from tests.utils import (
    append_stream_to_file,
    stream_multiple,
    stream_with_params,
    create_messages,
    append_message_to_file,
)


class TestMainLoopByReadingDltFile(unittest.TestCase):
    def setUp(self):
        # Empty content dlt file is created
        _, self.dlt_file_name = tempfile.mkstemp(suffix=b".dlt")
        self.dlt_reader = cDLTFile(filename=self.dlt_file_name, is_live=True, iterate_unblock_mode=False)
        # message_queue to store the dispatched messages from main loop
        self.message_queue = []
        # When callback() is called, then it is reset to True
        self.callback_is_called = False
        # With this variable, we could test different return value from callback()
        # If callback() returns True, then main loop keeps going; otherwise, it breaks
        self.callback_return_value = True
        # Thread for main loop, which is instantiated in test case
        self.main_loop = None

    def _callback_for_message(self, message):
        self.callback_is_called = True
        print("Called here")
        if message:
            self.message_queue.append(message)
        return self.callback_return_value

    def _start_main_loop(self):
        self.main_loop = Thread(
            target=py_dlt_file_main_loop,
            kwargs={"dlt_reader": self.dlt_reader, "callback": self._callback_for_message},
        )
        # self.main_loop.daemon = True
        self.main_loop.start()
        time.sleep(1)

    def tearDown(self):
        if not self.dlt_reader.stop_reading_proc.is_set():
            self.dlt_reader.stop_reading_proc.set()
            # After the stop of dlt_reader, main loop should be stopped automatically
            if self.main_loop:
                for _ in range(5):
                    if not self.main_loop.is_alive():
                        break
                    time.sleep(0.1)
                self.assertFalse(self.main_loop.is_alive())
        os.remove(self.dlt_file_name)

    def test_001_empty_dlt_file(self):
        """When dlt file has empty content, then no message could be dispatched, and no return value from main loop"""
        self._start_main_loop()
        time.sleep(0.1)
        # When file has empty content, callback() will not be called by any message
        self.assertFalse(self.callback_is_called)
        self.assertEqual(0, len(self.message_queue))

    def test_002_first_write_then_read_dlt_file(self):
        """
        Simulate a real dlt file case: first write to it, and then use main loop to read it
        """
        # First write to dlt file without opening main loop
        append_stream_to_file(stream_multiple, self.dlt_file_name)
        time.sleep(0.1)
        # Expectation: py_dlt_file_main_loop reads out the first batch messages to message_queue
        self._start_main_loop()
        time.sleep(0.1)
        self.assertTrue(self.callback_is_called)
        self.assertEqual(2, len(self.message_queue))

    def test_003_first_read_then_write_dlt_file(self):
        """
        Simulate a real dlt file case: first open main loop to read, then write to the file at opening main loop
        """
        # First only main loop to read dlt file
        self._start_main_loop()
        # Then write to dlt file
        append_stream_to_file(stream_multiple, self.dlt_file_name)
        time.sleep(0.1)
        # Expect the written logs could be dispatched by main loop
        self.assertTrue(self.callback_is_called)
        self.assertEqual(2, len(self.message_queue))

    def test_004_read_2_writes(self):
        """
        Test main loop reads from 2 consecutive writes to dlt file
        """
        # First only main loop to read dlt file
        self._start_main_loop()
        # First write to dlt file
        append_stream_to_file(stream_multiple, self.dlt_file_name)
        time.sleep(0.1)
        # Expect main loop could dispatch 2 logs
        self.assertTrue(self.callback_is_called)
        self.assertEqual(2, len(self.message_queue))
        # Second write to dlt file, and expect to dispatch 3 logs
        append_stream_to_file(stream_with_params, self.dlt_file_name)
        time.sleep(0.1)
        self.assertEqual(3, len(self.message_queue))

    def test_005_callback_return_false(self):
        """
        If callback returns false, then main loop should exit
        """
        # Set callback return value to False
        self.callback_return_value = False
        # Write to file
        append_stream_to_file(stream_multiple, self.dlt_file_name)
        time.sleep(0.1)
        # Open main loop to dispatch logs
        self._start_main_loop()
        # Expect main loop could dispatch 2 logs
        self.assertTrue(self.callback_is_called)
        # Callback returns False after it handles the first message, which terminates main loop
        # So, main loop wont be able to proceed the second message
        self.assertEqual(1, len(self.message_queue))
        self.assertFalse(self.main_loop.is_alive())

    def test_006_read_empty_apid_ctid_message(self):
        """
        Simulate a case to read a apid==b"" and ctid==b"" message
        """
        # Construct a message with apid==b"" and ctid==b""
        message = create_messages(stream_with_params, from_file=True)[0]
        message.extendedheader.apid = b""
        message.extendedheader.ctid = b""
        # Write this message into dlt file
        append_message_to_file(message, self.dlt_file_name)
        # Expectation: py_dlt_file_main_loop reads out the first batch messages to message_queue
        self._start_main_loop()
        time.sleep(0.1)
        self.assertTrue(self.callback_is_called)
        self.assertEqual(1, len(self.message_queue))
        # Expectation: the received message should have apid==b"" and ctid==b""
        self.assertEqual("", self.message_queue[0].apid)
        self.assertEqual("", self.message_queue[0].ctid)



# Copyright (C) 2017. BMW Car IT GmbH. All rights reserved.
"""Basic size tests for ctype wrapper definitions, to protect against regressions"""

import ctypes
import importlib
import os
import unittest
from unittest.mock import patch, MagicMock

import dlt


class TestCoreStructures(unittest.TestCase):
    def setUp(self):
        self.size_map = {
            "cDltServiceConnectionInfo": 10,
            "cDltStorageHeader": 16,
            "cDltStandardHeader": 4,
            "cDltStandardHeaderExtra": 12,
            "cDltExtendedHeader": 10,
            "cDLTMessage": 120,
            "cDltReceiver": 72,
            "cDltClient": 144,
            "cDLTFilter": 604,
        }

    def test_sizeof(self):
        importlib.import_module("dlt.core")

        api_ver = dlt.core.get_version(dlt.core.dltlib)
        ver_info = tuple(int(num) for num in api_ver.split("."))
        if ver_info >= (3, 0, 0):
            self.size_map["cDLTMessage"] = 103
            self.size_map["cDltClient"] = 152
            self.size_map["cDLTFilter"] = 1152

        for clsname, expected in self.size_map.items():
            actual = ctypes.sizeof(getattr(dlt.core, clsname))
            self.assertEqual(
                actual,
                expected,
                msg="v{0}, sizeof {1}: {2} != {3}".format(
                    dlt.core.get_version(dlt.core.dltlib), clsname, actual, expected
                ),
            )


class TestImportSpecificVersion(unittest.TestCase):
    def setUp(self):
        self.original_api_version = dlt.core.API_VER
        self.version_answer = b"2.18.5"
        self.version_str = (
            b"DLT Package Version: 2.18.5 STABLE, Package Revision: v2.18.5_5_g33fbad1, "
            b"build on Sep  2 2020 11:55:50\n-SYSTEMD -SYSTEMD_WATCHDOG -TEST -SHM\n"
        )
        self.version_filename = "core_2185.py"
        self.version_truncate_str = "2.18.5"
        self.version_truncate_filename = "core_2180.py"

        dlt.core.API_VER = None

    def tearDown(self):
        dlt.core.API_VER = self.original_api_version

    def test_get_version(self):
        def mock_dlt_get_version(buf, buf_size):
            ver_cstr = ctypes.create_string_buffer(self.version_str)
            ctypes.memmove(buf, ver_cstr, len(ver_cstr))

        mock_loaded_lib = MagicMock()
        mock_loaded_lib.dlt_get_version = MagicMock(side_effect=mock_dlt_get_version)

        api_version = dlt.core.get_version(mock_loaded_lib)
        self.assertEqual(mock_loaded_lib.dlt_get_version.call_count, 1)

        self.assertEqual(api_version, self.version_answer.decode())
        self.assertEqual(dlt.core.API_VER, self.version_answer.decode())

    def test_get_api_specific_file(self):
        with patch.object(os.path, "exists", return_value=True):
            filename = dlt.core.get_api_specific_file(self.version_answer.decode())
            self.assertEqual(filename, self.version_filename)

    def test_get_api_specific_file_not_found(self):
        with patch.object(os.path, "exists", return_value=False):
            with self.assertRaises(ImportError) as err_cm:
                dlt.core.get_api_specific_file(self.version_answer.decode())

            self.assertEqual(str(err_cm.exception), "No module file: core_200.py")

    def test_get_api_specific_file_truncate_minor_version(self):
        with patch.object(os.path, "exists", side_effect=[False, True]):
            filename = dlt.core.get_api_specific_file(self.version_truncate_str)
            self.assertEqual(filename, self.version_truncate_filename)



# Copyright (C) 2017. BMW Car IT GmbH. All rights reserved.
"""DLT Receive using py_dlt"""

import argparse
import logging
import time

from dlt.dlt import DLT_DAEMON_TCP_PORT, DLT_UDP_MULTICAST_FD_BUFFER_SIZE, DLT_UDP_MULTICAST_BUFFER_SIZE
from dlt.dlt_broker import DLTBroker

logging.basicConfig(format="%(asctime)s %(name)s %(levelname)-8s %(message)s")
root_logger = logging.getLogger()  # pylint: disable=invalid-name
logger = logging.getLogger("py-dlt-receive")  # pylint: disable=invalid-name


def parse_args():
    """Parse command line arguments"""
    logger.info("Parsing arguments")
    parser = argparse.ArgumentParser(description="Receive DLT messages")
    parser.add_argument("--host", required=True, help="hostname or ip address to connect to")
    parser.add_argument(
        "--port",
        default=DLT_DAEMON_TCP_PORT,
        type=int,
        help=f"Port of the DLT Daemon to connect to. default: {DLT_DAEMON_TCP_PORT}",
    )
    parser.add_argument("--file", required=True, help="The file into which the messages will be written")
    parser.add_argument(
        "--udp-fd-buffer-size",
        dest="udp_fd_buffer_size",
        default=DLT_UDP_MULTICAST_FD_BUFFER_SIZE,
        type=int,
        help=f"Set the socket buffer size in udp multicast mode. default: {DLT_UDP_MULTICAST_FD_BUFFER_SIZE} bytes",
    )
    parser.add_argument(
        "--udp-buffer-size",
        dest="udp_buffer_size",
        default=DLT_UDP_MULTICAST_BUFFER_SIZE,
        type=int,
        help=f"Set the DltReceiver buffer size in udp multicast mode. default: {DLT_UDP_MULTICAST_BUFFER_SIZE} bytes",
    )
    return parser.parse_args()


def dlt_receive(options):
    """Receive DLT messages via DLTBroker"""
    logger.info("Creating DLTBroker instance")
    broker = DLTBroker(
        ip_address=options.host,
        port=options.port,
        filename=options.file,
        udp_fd_buffer_size_bytes=options.udp_buffer_size,
        udp_buffer_size_bytes=options.udp_fd_buffer_size,
    )

    logger.info("Starting DLTBroker")
    broker.start()  # start the loop
    try:
        logger.info("Receiving messages...")
        while True:
            time.sleep(0.1)
    except KeyboardInterrupt:
        logger.info("Interrupted...")
    finally:
        logger.info("Stopping DLT broker")
        broker.stop()
        logger.info("Stopped DLT broker")


def main():
    """Main function"""
    root_logger.setLevel(level=logging.INFO)

    options = parse_args()
    logger.info("Parsed arguments: %s", options)

    dlt_receive(options)


if __name__ == "__main__":
    main()



# Copyright (C) 2022. BMW CTW PT. All rights reserved.
"""v2.18.8 specific class definitions"""

import ctypes
import logging

from dlt.core.core_base import dltlib

# DltClientMode from dlt_client.h
DLT_CLIENT_MODE_UNDEFINED = -1
DLT_CLIENT_MODE_TCP = 0
DLT_CLIENT_MODE_SERIAL = 1
DLT_CLIENT_MODE_UNIX = 2
DLT_CLIENT_MODE_UDP_MULTICAST = 3

# DltReceiverType from dlt_common.h
DLT_RECEIVE_SOCKET = 0
DLT_RECEIVE_UDP_SOCKET = 1
DLT_RECEIVE_FD = 2
DLT_ID_SIZE = 4
DLT_FILTER_MAX = 30  # Maximum number of filters
DLT_RETURN_ERROR = -1

# Return value for DLTFilter.add() - exceeded maximum number of filters
MAX_FILTER_REACHED = 1
# Return value for DLTFilter.add() - specified filter already exists
REPEATED_FILTER = 2

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class sockaddr_in(ctypes.Structure):  # pylint: disable=invalid-name
    """Auxiliary definition for cDltReceiver. Defined in netinet/in.h header"""

    _fields_ = [
        ("sa_family", ctypes.c_ushort),  # sin_family
        ("sin_port", ctypes.c_ushort),
        ("sin_addr", ctypes.c_byte * 4),
        ("__pad", ctypes.c_byte * 8),
    ]  # struct sockaddr_in is 16


class cDltReceiver(ctypes.Structure):  # pylint: disable=invalid-name
    """The structure is used to organise the receiving of data including buffer handling.
    This structure is used by the corresponding functions.

    typedef struct
    {
        int32_t lastBytesRcvd;    /**< bytes received in last receive call */
        int32_t bytesRcvd;        /**< received bytes */
        int32_t totalBytesRcvd;   /**< total number of received bytes */
        char *buffer;             /**< pointer to receiver buffer */
        char *buf;                /**< pointer to position within receiver buffer */
        char *backup_buf;         /** pointer to the buffer with partial messages if any **/
        int fd;                   /**< connection handle */
        DltReceiverType type;     /**< type of connection handle */
        int32_t buffersize;       /**< size of receiver buffer */
        struct sockaddr_in addr;  /**< socket address information */
    } DltReceiver;
    """

    _fields_ = [
        ("lastBytesRcvd", ctypes.c_int32),
        ("bytesRcvd", ctypes.c_int32),
        ("totalBytesRcvd", ctypes.c_int32),
        ("buffer", ctypes.POINTER(ctypes.c_char)),
        ("buf", ctypes.POINTER(ctypes.c_char)),
        ("backup_buf", ctypes.POINTER(ctypes.c_char)),
        ("fd", ctypes.c_int),
        ("type", ctypes.c_int),
        ("buffersize", ctypes.c_int32),
        ("addr", sockaddr_in),
    ]


class cDltClient(ctypes.Structure):  # pylint: disable=invalid-name
    """
    typedef struct
    {
        DltReceiver receiver;      /**< receiver pointer to dlt receiver structure */
        int sock;                  /**< sock Connection handle/socket */
        char *servIP;              /**< servIP IP adress/Hostname of TCP/IP interface */
        char *hostip;              /**< IP multicast address of group */
        int port;                  /**< Port for TCP connections (optional) */
        char *serialDevice;        /**< serialDevice Devicename of serial device */
        char *socketPath;          /**< socketPath Unix socket path */
        char ecuid[4];             /**< ECUiD */
        speed_t baudrate;          /**< baudrate Baudrate of serial interface, as speed_t */
        DltClientMode mode;        /**< mode DltClientMode */
        int send_serial_header;    /**< (Boolean) Send DLT messages with serial header */
        int resync_serial_header;  /**< (Boolean) Resync to serial header on all connection */
    } DltClient;
    """

    _fields_ = [
        ("receiver", cDltReceiver),
        ("sock", ctypes.c_int),
        ("servIP", ctypes.c_char_p),
        ("hostip", ctypes.c_char_p),
        ("port", ctypes.c_int),
        ("serialDevice", ctypes.c_char_p),
        ("socketPath", ctypes.c_char_p),
        ("ecuid", ctypes.c_char * 4),
        ("baudrate", ctypes.c_uint),
        ("mode", ctypes.c_int),
        ("send_serial_header", ctypes.c_int),
        ("resync_serial_header", ctypes.c_int),
    ]


class cDLTFilter(ctypes.Structure):  # pylint: disable=invalid-name
    """
    typedef struct
    {
        char apid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< application id */
        char ctid[DLT_FILTER_MAX][DLT_ID_SIZE]; /**< context id */
        int log_level[DLT_FILTER_MAX];          /**< log level */
        int32_t payload_max[DLT_FILTER_MAX];    /**< upper border for payload */
        int32_t payload_min[DLT_FILTER_MAX];    /**< lower border for payload */
        int  counter;                           /**< number of filters */
    } DltFilter;
    """

    _fields_ = [
        ("apid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
        ("ctid", (ctypes.c_char * DLT_ID_SIZE) * DLT_FILTER_MAX),
        ("log_level", ctypes.c_int * DLT_FILTER_MAX),
        ("payload_max", (ctypes.c_int32 * DLT_FILTER_MAX)),
        ("payload_min", (ctypes.c_int32 * DLT_FILTER_MAX)),
        ("counter", ctypes.c_int),
    ]

    # pylint: disable=too-many-arguments
    def add(self, apid, ctid, log_level=0, payload_min=0, payload_max=ctypes.c_uint32(-1).value // 2):
        """Add new filter pair"""
        if isinstance(apid, str):
            apid = bytes(apid, "ascii")
        if isinstance(ctid, str):
            ctid = bytes(ctid, "ascii")
        if (
            dltlib.dlt_filter_add(
                ctypes.byref(self), apid or b"", ctid or b"", log_level, payload_min, payload_max, self.verbose
            )
            == DLT_RETURN_ERROR
        ):
            if self.counter >= DLT_FILTER_MAX:
                logger.error("Maximum number (%d) of allowed filters reached, ignoring filter!\n", DLT_FILTER_MAX)
                return MAX_FILTER_REACHED
            logger.debug("Filter ('%s', '%s') already exists", apid, ctid)
            return REPEATED_FILTER
        return 0



# Copyright (C) 2015. BMW Car IT GmbH. All rights reserved.
"""Pure Python implementation of DLT library"""

import ctypes
import ipaddress as ip
import logging
import os
import re
import socket
import struct
import sys
import time
import threading
import multiprocessing

from dlt.core import (
    cDLTFilter,
    dltlib,
    DLT_CLIENT_MODE_UDP_MULTICAST,
    DLT_ID_SIZE,
    DLT_HTYP_WEID,
    DLT_HTYP_WSID,
    DLT_HTYP_WTMS,
    DLT_HTYP_UEH,
    DLT_RETURN_OK,
    DLT_RETURN_ERROR,
    DLT_RETURN_TRUE,
    DLT_MESSAGE_ERROR_OK,
    cDltExtendedHeader,
    cDltClient,
    MessageMode,
    cDLTMessage,
    cDltStorageHeader,
    cDltStandardHeader,
    DLT_TYPE_INFO_BOOL,
    DLT_TYPE_INFO_UINT,
    DLT_TYPE_INFO_SINT,
    DLT_TYPE_INFO_FLOA,
    DLT_TYPE_INFO_STRG,
    DLT_TYPE_INFO_SCOD,
    DLT_TYPE_INFO_TYLE,
    DLT_TYPE_INFO_VARI,
    DLT_TYPE_INFO_RAWD,
    DLT_SCOD_ASCII,
    DLT_SCOD_UTF8,
    DLT_TYLE_8BIT,
    DLT_TYLE_16BIT,
    DLT_TYLE_32BIT,
    DLT_TYLE_64BIT,
    DLT_TYLE_128BIT,
    DLT_DAEMON_TCP_PORT,
    DLT_CLIENT_RCVBUFSIZE,
    DLT_RECEIVE_SOCKET,
)
from dlt.helpers import bytes_to_str

MAX_LOG_IN_ROW = 3
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

DLT_EMPTY_FILE_ERROR = "DLT TRACE FILE IS EMPTY"
cDLT_FILE_NOT_OPEN_ERROR = "Could not open DLT Trace file (libdlt)"  # pylint: disable=invalid-name

DLT_UDP_MULTICAST_FD_BUFFER_SIZE = int(os.environ.get("PYDLT_UDP_MULTICAST_FD_BUFFER_SIZE", 2 * (2**20)))  # 2 Mb
DLT_UDP_MULTICAST_BUFFER_SIZE = int(os.environ.get("PYDLT_UDP_MULTICAST_BUFFER_SIZE", 8 * (2**20)))  # 8 Mb


class cached_property(object):  # pylint: disable=invalid-name
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Copyright: Marcel Hellkamp <marc@gsites.de>
    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
    Licence: MIT
    """  # noqa

    def __init__(self, func):
        self.__doc__ = getattr(func, "__doc__")
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


class DLTFilter(cDLTFilter):
    """Structure to store filter parameters. ID are maximal four characters. Unused values are filled with zeros.
    If every value as filter is valid, the id should be empty by having only zero values.
    """

    verbose = 0

    def __init__(self, **kwords):
        self.verbose = kwords.pop("verbose", 0)
        if dltlib.dlt_filter_init(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
            raise RuntimeError("Could not initialize DLTFilter")
        super(DLTFilter, self).__init__(**kwords)

    def __del__(self):
        if dltlib.dlt_filter_free(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
            raise RuntimeError("Could not cleanup DLTFilter")

    def __repr__(self):
        """Return the 'official' string representation of an object"""
        apids = [ctypes.string_at(entry[:DLT_ID_SIZE]) for entry in self.apid]
        ctids = [ctypes.string_at(entry[:DLT_ID_SIZE]) for entry in self.ctid]

        return str(list(zip(apids[: self.counter], ctids[: self.counter])))

    def __nonzero__(self):
        """Truth value testing"""
        return self.counter > 0

    __bool__ = __nonzero__


class Payload(object):
    """Payload object encapsulates the payload decoding and list-like access to arguments"""

    def __init__(self, message):
        self._params = None
        self._noar = message.noar
        self._buf = ctypes.string_at(message.databuffer, message.datasize)

    def __getitem__(self, index):
        """Accessing the payload item as a list"""
        if index < 0 or index > self._noar:
            return IndexError()

        # we have parsed it already - just return the item
        if self._params is not None:
            return self._params[index]

        self._parse_payload()

        return self._params[index]

    def _parse_payload(self):  # pylint: disable=too-many-branches,too-many-statements
        """Parse the payload into list of arguments"""
        self._params = []

        offset = 0
        for _ in range(self._noar):
            type_info = struct.unpack_from("I", self._buf, offset)[0]
            offset += struct.calcsize("I")

            def get_scod(type_info):
                """Helper function"""
                return type_info & DLT_TYPE_INFO_SCOD

            value = None
            if type_info & DLT_TYPE_INFO_STRG:
                if (get_scod(type_info) == DLT_SCOD_ASCII) or (get_scod(type_info) == DLT_SCOD_UTF8):
                    length = struct.unpack_from("H", self._buf, offset)[0]
                    offset += struct.calcsize("H")
                    value = self._buf[offset : offset + length - 1]  # strip the string terminating char \x00
                    offset += length

            elif type_info & DLT_TYPE_INFO_UINT:
                if type_info & DLT_TYPE_INFO_VARI:
                    pass

                tyle = type_info & DLT_TYPE_INFO_TYLE
                if tyle == DLT_TYLE_8BIT:
                    value = struct.unpack_from("B", self._buf, offset)[0]
                    offset += 1
                elif tyle == DLT_TYLE_16BIT:
                    value = struct.unpack_from("H", self._buf, offset)[0]
                    offset += 2
                elif tyle == DLT_TYLE_32BIT:
                    value = struct.unpack_from("I", self._buf, offset)[0]
                    offset += 4
                elif tyle == DLT_TYLE_64BIT:
                    value = struct.unpack_from("Q", self._buf, offset)[0]
                    offset += 8
                elif tyle == DLT_TYLE_128BIT:
                    value = struct.unpack_from("QQ", self._buf, offset)
                    offset += 16
                    if sys.byteorder == "little":
                        value = (value[1] << 64) | value[0]
                    else:
                        value = (value[0] << 64) | value[1]

            elif type_info & DLT_TYPE_INFO_SINT:
                if type_info & DLT_TYPE_INFO_VARI:
                    pass

                tyle = type_info & DLT_TYPE_INFO_TYLE
                if tyle == DLT_TYLE_8BIT:
                    value = struct.unpack_from("b", self._buf, offset)[0]
                    offset += 1
                elif tyle == DLT_TYLE_16BIT:
                    value = struct.unpack_from("h", self._buf, offset)[0]
                    offset += 2
                elif tyle == DLT_TYLE_32BIT:
                    value = struct.unpack_from("i", self._buf, offset)[0]
                    offset += 4
                elif tyle == DLT_TYLE_64BIT:
                    value = struct.unpack_from("q", self._buf, offset)[0]
                    offset += 8
                elif tyle == DLT_TYLE_128BIT:
                    value = struct.unpack_from("qq", self._buf, offset)
                    offset += 16
                    if sys.byteorder == "little":
                        value = (value[1] << 64) | value[0]
                    else:
                        value = (value[0] << 64) | value[1]

            elif type_info & DLT_TYPE_INFO_FLOA:
                if type_info & DLT_TYLE_64BIT:
                    value = struct.unpack_from("d", self._buf, offset)[0]
                    offset += 8
                else:
                    value = struct.unpack_from("f", self._buf, offset)[0]
                    offset += 4

            elif type_info & DLT_TYPE_INFO_RAWD:
                if type_info & DLT_TYPE_INFO_VARI:
                    pass

                length = struct.unpack_from("H", self._buf, offset)[0]
                offset += struct.calcsize("H")

                value = self._buf[offset : offset + length]
                offset += length

            elif type_info & DLT_TYPE_INFO_BOOL:
                value = struct.unpack_from("?", self._buf, offset)[0]
                offset += 1

            else:
                value = "ERROR"

            self._params.append(value)

    def __len__(self):
        """Return number of parsed parameters"""
        if self._params is None:
            self._parse_payload()

        return len(self._params)


class DLTMessage(cDLTMessage, MessageMode):
    """Python wrapper class for the cDLTMessage structure"""

    verbose = 0

    # object is not initialized if the message is loaded from a file
    initialized_as_object = False

    re_pattern_type = type(re.compile(r"type"))

    def __init__(self, *args, **kwords):
        self.initialized_as_object = True
        self.verbose = kwords.pop("verbose", 0)
        if self.verbose:
            logger.debug("DLTMessage._init_(%s)", kwords)
        self.lifecycle = None

        if dltlib.dlt_message_init(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
            raise RuntimeError("Could not initialize DLTMessage")

        super(DLTMessage, self).__init__(*args, **kwords)

    def __reduce__(self):
        """Pickle serialization API

        This method is called by the pickle module to serialize objects
        that it cannot automatically serialize.
        """
        # copy the data from the databuffer pointer into an array
        databuffer = ctypes.ARRAY(ctypes.c_uint8, self.datasize)()
        ctypes.memmove(databuffer, self.databuffer, self.datasize)

        init_args = (self.found_serialheader, self.resync_offset, self.headersize, self.datasize)
        state_dict = {
            "headerbuffer": bytearray(self.headerbuffer),
            "databuffer": bytearray(databuffer),
            "databuffersize": self.databuffersize,
            "storageheader": self.storageheader,
            "standardheader": self.standardheader,
            "headerextra": self.headerextra,
            "extendedheader": self.extendedheader,
        }
        return (DLTMessage, init_args, state_dict)

    # pylint: disable=attribute-defined-outside-init
    def __setstate__(self, state):
        """Pickle deserialization API

        This method is called by the pickle module to populate a
        deserialized object's state after it has been created.
        """
        self.databuffersize = state["databuffersize"]
        self.p_storageheader.contents = state["storageheader"]
        self.p_standardheader.contents = state["standardheader"]
        self.headerextra = state["headerextra"]
        self.p_extendedheader.contents = state["extendedheader"]
        # - populate databuffer
        databuffer = ctypes.ARRAY(ctypes.c_uint8, self.datasize)()
        for index, byte in enumerate(state["databuffer"]):
            databuffer[index] = byte
        self.databuffer = databuffer

        # - populate headerbuffer
        for index, byte in enumerate(state["headerbuffer"]):
            self.headerbuffer[index] = byte

        # - This is required because we are not calling
        # dlt_message_init() so we do not need to call
        # dlt_message_free()
        self.initialized_as_object = False

    @staticmethod
    def from_bytes(data):
        """Create a class instance from a byte string in DLT storage format"""
        msg = DLTMessage()
        storageheader, remainder = msg.extract_storageheader(data)

        buf = ctypes.create_string_buffer(remainder)

        dltlib.dlt_message_read(
            ctypes.byref(msg),
            ctypes.cast(buf, ctypes.POINTER(ctypes.c_uint8)),
            ctypes.c_uint(len(remainder)),
            0,  # resync
            0,
        )  # verbose
        msg.p_storageheader.contents = storageheader
        msg.initialized_as_object = False

        return msg

    def to_bytes(self):
        """Create DLT storage format bytes from DLTMessage instance"""
        return ctypes.string_at(self.headerbuffer, self.headersize) + ctypes.string_at(self.databuffer, self.datasize)

    def __copy__(self):
        """Create a copy of the message"""
        return DLTMessage.from_bytes(self.to_bytes())

    @staticmethod
    def extract_storageheader(data):
        """Split binary message data into storage header and remainder"""
        header = data[0 : ctypes.sizeof(cDltStorageHeader)]
        # pylint: disable=no-member
        return (cDltStorageHeader.from_buffer_copy(header), data[ctypes.sizeof(cDltStorageHeader) :])

    @staticmethod
    def extract_sort_data(data):
        """Extract timestamp, message length, apid, ctid from a bytestring in DLT storage format (speed optimized)"""
        htyp_data = ord(chr(data[16]))
        len_data = data[19:17:-1]
        len_value = ctypes.cast(len_data, ctypes.POINTER(ctypes.c_ushort)).contents.value + 16
        apid = b""
        ctid = b""
        tmsp_value = 0.0

        bytes_offset = 0  # We know where data will be in the message, but ...
        if not htyp_data & DLT_HTYP_WEID:  # if there is no ECU ID and/or Session ID, then it will be earlier
            bytes_offset -= 4
        if not htyp_data & DLT_HTYP_WSID:
            bytes_offset -= 4

        if htyp_data & DLT_HTYP_WTMS:
            tmsp_base = 31 + bytes_offset  # Typical timestamp end offset
            tmsp_data = data[tmsp_base : tmsp_base - 4 : -1]
            tmsp_value = ctypes.cast(tmsp_data, ctypes.POINTER(ctypes.c_uint32)).contents.value / 10000.0

        if htyp_data & DLT_HTYP_UEH:
            apid_base = 38 + bytes_offset  # Typical APID end offset
            apid = data[apid_base - 4 : apid_base].rstrip(b"\x00")
            ctid = data[apid_base : apid_base + 4].rstrip(b"\x00")

        apid = bytes_to_str(apid)
        ctid = bytes_to_str(ctid)

        return tmsp_value, len_value, apid, ctid

    def __del__(self):
        if self.initialized_as_object is True:
            if dltlib.dlt_message_free(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
                raise RuntimeError("Could not free DLTMessage")

    @property
    def storageheader(self):
        """Workaround to get rid of need to call .contents"""
        try:
            return self.p_storageheader.contents
        except ValueError:
            return None

    @property
    def standardheader(self):
        """Workaround to get rid of need to call .contents"""
        return self.p_standardheader.contents

    @property
    def extendedheader(self):
        """Workaround to get rid of need to call .contents"""
        try:
            return self.p_extendedheader.contents
        except ValueError:
            return None

    def __eq__(self, other):
        """Equal test - not comparing storage header (contains timestamps)"""
        header1 = ctypes.string_at(self.headerbuffer, self.headersize)[ctypes.sizeof(cDltStorageHeader) :]
        header2 = ctypes.string_at(other.headerbuffer, other.headersize)[ctypes.sizeof(cDltStorageHeader) :]

        data1 = ctypes.string_at(self.databuffer, self.datasize)
        data2 = ctypes.string_at(other.databuffer, other.datasize)

        return header1 == header2 and data1 == data2

    def compare(self, other=None):  # pylint: disable=too-many-return-statements,too-many-branches
        """Compare messages by given attributes

        :param [DLTMessage|DLTFilter|dict] other: DLTMessage object (or DLTFilter or a dict with selected keys)
            to compare with. Use DLTFilter object with APID,CTID pairs for the best performance.
        :returns: True if all attributes match or False if any of the given attributes differs
        :rtype: bool
        :raises TypeError: if other is neither DLTMessage nor a dictionary

        Example:
        message.compare(other=message2)
        message.compare(message2)
        message.compare(other=dict(apid="AP1", ctid="CT1"))
        message.compare(dict(apid="AP1", ctid="CT1"))
        message.compare(dict(apid=re.compile(r"^A.*"))  # match all messages which apid starting with A
        message.compare(dict(apid="AP1", ctid="CT1", payload_decoded=re.compile(r".connected.*")))
        """
        if hasattr(other, "apid") and hasattr(other, "ctid") and hasattr(other, "payload_decoded"):
            # other is DLTMessage - full compare
            return self.apid == other.apid and self.ctid == other.ctid and self.__eq__(other)

        # pylint: disable=protected-access
        if hasattr(other, "_fields_") and hasattr(other, "counter") and hasattr(other, "apid"):
            # other is DLTFilter or cDLTFilter
            return dltlib.dlt_message_filter_check(ctypes.byref(self), ctypes.byref(other), 0)

        if not isinstance(other, dict):
            raise TypeError(
                "other must be instance of dlt.dlt.DLTMessage, dlt.dlt.DLTFilter or a dictionary"
                " found: {}".format(type(other))
            )

        other = other.copy()
        apid = other.get("apid", None)
        if apid and not isinstance(apid, self.re_pattern_type) and self.apid != apid:
            return False

        ctid = other.get("ctid", None)
        if ctid and not isinstance(ctid, self.re_pattern_type) and self.ctid != ctid:
            return False

        for key, val in other.items():
            if val is None:
                continue
            key = key.rsplit(".", 1)[-1]  # In case the obsolete "extendedheader.apid" notation is used
            msg_val = getattr(self, key, b"")
            if not msg_val:
                return False
            if isinstance(val, self.re_pattern_type):
                if not val.search(msg_val):
                    return False
            elif msg_val != val:
                return False
        return True

    def __str__(self):
        """Construct DLTViewer-like string"""
        out = [time.asctime(time.gmtime(self.storage_timestamp))]
        if self.headerextra:
            out.append(self.headerextra.tmsp / 10000.0)
        out += [self.standardheader.mcnt, self.storageheader.ecu]
        if self.extendedheader:
            out += [self.extendedheader.apid, self.extendedheader.ctid]
        if self.headerextra:
            out.append(self.headerextra.seid)
        out += [self.type_string, self.subtype_string, self.mode_string, self.noar, self.payload_decoded]
        return " ".join(bytes_to_str(item) for item in out)

    # convenient access to import DLT message attributes
    # no need to remember in which header are those attrs defined
    @cached_property
    def ecuid(self):  # pylint: disable=invalid-overridden-method
        """Get the ECU ID

        :returns: ECU ID
        :rtype: str
        """
        return bytes_to_str(self.storageheader.ecu or self.headerextra.ecu)

    @cached_property
    def mcnt(self):  # pylint: disable=invalid-overridden-method
        """Get the message counter index

        :returns: message index
        :rtype: int
        """
        return int(self.standardheader.mcnt)

    @cached_property
    def seid(self):  # pylint: disable=invalid-overridden-method
        """Get the Session ID if WSID is set in the message type, otherwise 0

        :returns: Session ID
        :rtype: int
        """
        return int(self.headerextra.seid) if (self.standardheader.htyp & DLT_HTYP_WSID) else 0

    @cached_property
    def tmsp(self):  # pylint: disable=invalid-overridden-method
        """Get the timestamp

        :returns: timestamp
        :rtype: float [s]
        """
        return (self.headerextra.tmsp / 10000.0) if (self.standardheader.htyp & DLT_HTYP_WTMS) else 0

    @cached_property
    def apid(self):  # pylint: disable=invalid-overridden-method
        """Get the Application ID

        :returns: Application ID
        :rtype: str
        """
        return bytes_to_str(self.extendedheader.apid if self.extendedheader else "")

    @cached_property
    def ctid(self):  # pylint: disable=invalid-overridden-method
        """Get the Context ID

        :returns: Context ID
        :rtype: str
        """
        return bytes_to_str(self.extendedheader.ctid if self.extendedheader else "")

    @cached_property
    def noar(self):  # pylint: disable=invalid-overridden-method
        """Get the number of arguments

        :returns: Context ID
        :rtype: str
        """
        if self.use_extended_header and self.is_mode_verbose:
            return self.extendedheader.noar
        return 0

    @cached_property
    def payload(self):  # pylint: disable=invalid-overridden-method
        """Get the payload object

        :returns: Payload object
        :rtype: Payload
        """
        return Payload(self)

    @cached_property
    def payload_decoded(self):  # pylint: disable=invalid-overridden-method
        """Get the payload string

        :returns: Payload string
        :rtype: str
        """
        return bytes_to_str(super(DLTMessage, self).payload_decoded)

    @cached_property
    def storage_timestamp(self):  # pylint: disable=invalid-overridden-method
        """Get the storage header timestamp in seconds

        :returns: storage header timestamp
        :rtype: float
        """
        return self.storageheader.seconds + self.storageheader.microseconds * 0.000001


class cDLTFile(ctypes.Structure):  # pylint: disable=invalid-name
    """The structure to organise the access to DLT files. This structure is used by the corresponding functions.

    typedef struct sDltFile
    {
        /* file handle and index for fast access */
        FILE *handle;      /**< file handle of opened DLT file */
        long *index;       /**< file positions of all DLT messages for fast access to file, only filtered messages */

        /* size parameters */
        int32_t counter;       /**< number of messages in DLT file with filter */
        int32_t counter_total; /**< number of messages in DLT file without filter */
        int32_t position;      /**< current index to message parsed in DLT file starting at 0 */
        uint64_t file_length;  /**< length of the file */
        uint64_t file_position; /**< current position in the file */

        /* error counters */
        int32_t error_messages; /**< number of incomplete DLT messages found during file parsing */

        /* filter parameters */
        DltFilter *filter;  /**< pointer to filter list. Zero if no filter is set. */
        int32_t filter_counter; /**< number of filter set */

        /* current loaded message */
        DltMessage msg;     /**< pointer to message */

    } DltFile;
    """

    _fields_ = [
        ("handle", ctypes.POINTER(ctypes.c_int)),
        ("index", ctypes.POINTER(ctypes.c_long)),
        ("counter", ctypes.c_int32),
        ("counter_total", ctypes.c_int32),
        ("position", ctypes.c_int32),
        ("file_length", ctypes.c_uint64),
        ("file_position", ctypes.c_uint64),
        ("error_messages", ctypes.c_int32),
        ("filter", ctypes.POINTER(DLTFilter)),
        ("filter_counter", ctypes.c_int32),
        ("msg", DLTMessage),
    ]

    def __init__(self, **kwords):
        self.verbose = kwords.pop("verbose", 0)
        self.filename = kwords.pop("filename", None)
        if isinstance(self.filename, str):
            self.filename = bytes(self.filename, "utf-8")
        super(cDLTFile, self).__init__(**kwords)
        if dltlib.dlt_file_init(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
            raise RuntimeError("Could not initialize DLTFile")
        self._iter_index = 0
        self.corrupt_msg_count = 0

        self.indexed = False
        self.end = False
        self.live_run = kwords.pop("is_live", False)
        # Stop event for threading usage in caller
        self.stop_reading = threading.Event()
        # Stop event for process usage in caller
        self.stop_reading_proc = multiprocessing.Event()

    def __repr__(self):
        # pylint: disable=bad-continuation
        return "<DLTFile object {} with {} messages>".format(
            "filename={}".format(self.filename) if self.filename else "", self.counter_total
        )

    def __del__(self):
        if dltlib.dlt_file_free(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
            raise RuntimeError("Could not cleanup DLTFile")

    def _find_next_header(self):
        """Helper function for generate_index to skip over invalid storage headers.

        :returns: Offset to the next storage header position (after
                  self.file_position), if it was found, or position of EOF if not
        :rtype: int
        """
        with open(self.filename, "rb") as fobj:
            last_position = self.file_position  # pylint: disable=access-member-before-definition
            fobj.seek(last_position)
            buf = fobj.read(1024)
            while buf:
                found = buf.find(b"DLT\x01")
                if found != -1:
                    return last_position + found
                last_position = fobj.tell()
                buf = fobj.read(1024)
        return None

    # pylint: disable=attribute-defined-outside-init,access-member-before-definition
    def generate_index(self):
        """Generate an index for the loaded DLT file

        :returns: True if file had been previously read and the index is
                  successfully generated, otherwise False
        :rtype: bool
        """
        if not self.filename:
            return False

        self.indexed = False
        if dltlib.dlt_file_open(ctypes.byref(self), self.filename, self.verbose) >= DLT_RETURN_OK:
            # load, analyse data file and create index list
            if self.file_length == 0:
                raise IOError(DLT_EMPTY_FILE_ERROR)
            while self.file_position < self.file_length:
                ret = dltlib.dlt_file_read(ctypes.byref(self), self.verbose)
                if ret < DLT_RETURN_OK:
                    # - This can happen if either the frame's storage
                    # header could not be read correctly or the frame is
                    # corrupt. If the frame's storage header could not
                    # be read correctly we try to get the next storage
                    # header and continue indexing
                    next_header_position = self._find_next_header()
                    if next_header_position:
                        if self.file_position == next_header_position:  # pylint: disable=no-else-break
                            # - This this implies that dltlib.dlt_read_file()
                            # returned due to an error other than invalid storage
                            # header because we already were at the correct
                            # header_position in the last iteration. So, we
                            # need to break out of the read/index loop.
                            break
                        else:
                            self.file_position = next_header_position
                            self.corrupt_msg_count += 1
                    else:
                        break
            self.indexed = True
        else:
            raise IOError(cDLT_FILE_NOT_OPEN_ERROR)
        return self.indexed

    def read(self, filename, filters=None):
        """Index the DLT trace file for optimized DLT Message access

        :param str filename: DLT log filename to read the messages from
        :param list filters: List of filters to apply [("APPID", "CTID"), ...]
        :returns: True if file was read and indexed successfully, otherwise False
        :rtype: bool
        """
        # load the filters
        self.set_filters(filters)

        if isinstance(filename, str):
            filename = bytes(filename, "utf-8")
        # read and index file
        self.filename = filename
        self.generate_index()
        return self.indexed

    def set_filters(self, filters):
        """Set filters to optimize access"""
        if filters is not None:
            dlt_filter = DLTFilter(verbose=self.verbose)
            for apid, ctid in filters:
                if isinstance(apid, str):
                    apid = bytes(apid, "ascii")
                if isinstance(ctid, str):
                    ctid = bytes(ctid, "ascii")
                dlt_filter.add(apid, ctid)
            self.filters = dlt_filter
            dltlib.dlt_file_set_filter(ctypes.byref(self), ctypes.byref(dlt_filter), self.verbose)

    def __getitem__(self, index):
        """Load a DLT message from opened file

        :param int index: Index of a message to load
        :returns: Loaded DLTMessage
        :rtype: DLTMessage object
        :raises IndexError: If message index is out of boundary
        """
        if index < 0:
            if self.counter == 0:
                self.read(self.filename)
            index = self.counter + index

        if index == 0 and self.counter == 0:
            self.read(self.filename)

        if index < 0 or index >= self.counter:
            raise IndexError("Index out of range (0 < %d < %d)" % (index, self.counter))

        dltlib.dlt_file_message(ctypes.byref(self), index, self.verbose)
        # deepcopy the object
        msg = DLTMessage.from_buffer_copy(self.msg)  # pylint: disable=no-member
        msg.databuffer.contents = ctypes.create_string_buffer(self.msg.datasize)
        ctypes.memmove(msg.databuffer, self.msg.databuffer, msg.datasize)

        # set the new storage header pointer
        offset = 0
        hdr = cDltStorageHeader.from_address(ctypes.addressof(msg.headerbuffer) + offset)  # pylint: disable=no-member
        msg.p_storageheader = ctypes.pointer(hdr)

        # set the new standard header pointer
        offset = ctypes.sizeof(cDltStorageHeader)
        hdr = cDltStandardHeader.from_address(ctypes.addressof(msg.headerbuffer) + offset)  # pylint: disable=no-member
        msg.p_standardheader = ctypes.pointer(hdr)
        # set the new extended header pointer
        if self.msg.use_extended_header:
            offset = ctypes.addressof(self.msg.p_extendedheader.contents) - ctypes.addressof(self.msg.headerbuffer)
            # pylint: disable=no-member
            hdr = cDltExtendedHeader.from_address(ctypes.addressof(msg.headerbuffer) + offset)
            msg.p_extendedheader = ctypes.pointer(hdr)

        return msg

    def _open_file(self):
        """Open the configured file for processing"""
        file_opened = False
        while not self._is_stop_reading_set():
            if dltlib.dlt_file_open(ctypes.byref(self), self.filename, self.verbose) >= DLT_RETURN_OK:
                file_opened = True
                break
            if not self.live_run:
                break
            time.sleep(0.5)

        if not file_opened:
            logger.error("DLT FILE OPEN FAILED - Analysis will not be performed")
            raise IOError(cDLT_FILE_NOT_OPEN_ERROR)

    def _log_message_progress(self):
        """Logs current message for progress information"""
        length = os.stat(self.filename).st_size
        logger.debug(
            "Processed %s messages (%s%% of %sfile) from %s, next message is apid %s, ctid %s",
            self.position,
            int(100 * self.file_position / length),
            "live " if self.live_run else "",
            self.filename,
            self.msg.apid,
            self.msg.ctid,
        )

    def _is_stop_reading_set(self):
        return self.stop_reading.is_set() or self.stop_reading_proc.is_set()

    def __iter__(self):  # pylint: disable=too-many-branches
        """Iterate over messages in the file"""
        logger.debug("Starting File Read")
        logger.debug(
            "File Position: %d File Counter: %d File Name: %s", self.file_position, self.counter, self.filename
        )
        cached_mtime = 0
        cached_file_pos = 0
        corruption_check_try = True

        self._open_file()

        found_data = False
        while not self._is_stop_reading_set() or corruption_check_try:  # pylint: disable=too-many-nested-blocks
            os_stat = os.stat(self.filename)
            mtime = os_stat.st_mtime

            if mtime != cached_mtime and os_stat.st_size or corruption_check_try:
                cached_mtime = mtime
                corruption_check_try = False

                while not self._is_stop_reading_set() and (
                    dltlib.dlt_file_read(ctypes.byref(self), self.verbose) >= DLT_RETURN_OK
                ):
                    found_data = True
                    if (
                        self.filter
                        and dltlib.dlt_message_filter_check(ctypes.byref(self.msg), self.filter, 0) != DLT_RETURN_TRUE
                    ):
                        continue

                    index = self.position
                    msg = self[index]
                    if not index % 100000:
                        self._log_message_progress()
                    yield msg

                if cached_file_pos != self.file_position:
                    # We were able to read messages, don't do a corrupt message check yet.
                    corruption_check_try = True
                    cached_file_pos = self.file_position
                else:
                    next_header_position = self._find_next_header()
                    if next_header_position:
                        if self.file_position == next_header_position:
                            if not self.live_run:
                                logger.warning("Incomplete message while parsing DLT file at %s", self.file_position)
                                break
                        else:
                            logger.warning("Found a corrupt message at %s, skipping it", self.file_position)
                            self.file_position = next_header_position
                            self.corrupt_msg_count += 1
                            corruption_check_try = True
                    # Wait for further messages to determine if corrupt, else just end of file
                    else:
                        if not self.live_run:
                            logger.info("End of file reached at %s", self.file_position)
                            break

            time.sleep(0.1)

        if not found_data:
            raise IOError(DLT_EMPTY_FILE_ERROR)

    def __len__(self):
        """Returns filtered file length"""
        return self.counter


class DLTClient(cDltClient):
    """DLTClient class takes care about correct initialization and cleanup"""

    verbose = 0

    def __init__(self, **kwords):
        """Initialize a DLTClient.

        :param servIP: Optional[str] - dlt server IP.
        :param hostIP: Optional[str] - Only available for udp multicast mode.
            Set host interface address.
        :param port: Optional[int] - dlt tcp daemon port.
        :param verbose: Optional[bool] - Enable verbose output.
        :param udp_fd_buffer_size_bytes: Optional[int] - Only available for udp
            multicast mode. Set the UDP buffer size through setsockopt (unit: bytes).
        :param udp_buffer_size_bytes: Optional[int] - Only available for udp
            multicast mode. Set the DltReceiver's buffer size (unit: bytes).
        """
        self.is_udp_multicast = False
        self.verbose = kwords.pop("verbose", 0)
        if dltlib.dlt_client_init(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
            raise RuntimeError("Could not initialize DLTClient")

        if "servIP" in kwords:
            serv_ip = kwords.pop("servIP")
            if isinstance(serv_ip, str):
                serv_ip = serv_ip.encode("utf8")
            ip_init_state = dltlib.dlt_client_set_server_ip(ctypes.byref(self), ctypes.create_string_buffer(serv_ip))
            if ip_init_state == DLT_RETURN_ERROR:
                raise RuntimeError("Could not initialize servIP for DLTClient")

            if ip.ip_address(serv_ip.decode("utf8")).is_multicast:
                logger.info("Initializing DLTClient using UDP")
                self.is_udp_multicast = True
                if "hostIP" in kwords:
                    host_ip = kwords.pop("hostIP")
                    if isinstance(host_ip, str):
                        host_ip = host_ip.encode("utf8")
                    ip_init_state = dltlib.dlt_client_set_host_if_address(
                        ctypes.byref(self), ctypes.create_string_buffer(host_ip)
                    )
                    if ip_init_state == DLT_RETURN_ERROR:
                        raise RuntimeError("Could not initialize multicast address for DLTClient")

                set_mode_state = dltlib.dlt_client_set_mode(ctypes.byref(self), DLT_CLIENT_MODE_UDP_MULTICAST)
                logger.info("DLTClient using UDP set mode state: %s", set_mode_state)
                if set_mode_state == DLT_RETURN_ERROR:
                    raise RuntimeError("Could not initialize socket mode for DLTClient")

        # attribute to hold a reference to the connected socket in case
        # we created a connection with a timeout (via python, as opposed
        # to dltlib). This avoids the socket object from being garbage
        # collected when it goes out of the connect() method scope
        self._connected_socket = None

        super(DLTClient, self).__init__(**kwords)

        # (re)set self.port, even for API version <2.16.0 since we use
        # it ourselves elsewhere
        self.port = kwords.get("port", DLT_DAEMON_TCP_PORT)

        self._udp_fd_buffer_size_bytes = kwords.get("udp_fd_buffer_size_bytes", DLT_UDP_MULTICAST_FD_BUFFER_SIZE)
        self._udp_buffer_size_bytes = kwords.get("udp_buffer_size_bytes", DLT_UDP_MULTICAST_BUFFER_SIZE)

    def connect(self, timeout=None):
        """Connect to the server

        If timeout is provided, block on connect until timeout occurs. If
        timeout is not provided or is None, try to connect and return
        immediately

        :param int|None timeout: Seconds to wait for connection
        :returns: True if connected successfully, False otherwise
        :rtype: bool
        """
        connected = None
        error_count = 0
        if not self.is_udp_multicast:
            if self.verbose:
                logger.info("Connecting DLTClient using TCP Connection")
            if timeout:
                end_time = time.time() + timeout
                while time.time() < end_time:
                    timeout_remaining = max(end_time - time.time(), 1) if timeout else None
                    try:
                        self._connected_socket = socket.create_connection(
                            (ctypes.string_at(self.servIP), self.port), timeout=timeout_remaining
                        )
                    except IOError as exc:
                        if error_count < MAX_LOG_IN_ROW and self.verbose:
                            logger.debug(
                                "DLT client connect failed to connect to %s:%s : %s", self.servIP, self.port, exc
                            )
                        error_count += 1
                        time.sleep(1)

                    if self._connected_socket:
                        # pylint: disable=attribute-defined-outside-init
                        self.sock = ctypes.c_int(self._connected_socket.fileno())
                        # - also init the receiver to replicate
                        # dlt_client_connect() behavior
                        if hasattr(self.receiver, "type"):
                            connected = dltlib.dlt_receiver_init(
                                ctypes.byref(self.receiver), self.sock, DLT_RECEIVE_SOCKET, DLT_CLIENT_RCVBUFSIZE
                            )
                        else:
                            connected = dltlib.dlt_receiver_init(
                                ctypes.byref(self.receiver), self.sock, DLT_CLIENT_RCVBUFSIZE
                            )
                        break
            else:
                connected = dltlib.dlt_client_connect(ctypes.byref(self), self.verbose)
                # - create a python socket object so that we can detect
                # connection loss in the main_loop below as described at
                # http://stefan.buettcher.org/cs/conn_closed.html
                self._connected_socket = socket.fromfd(self.sock, socket.AF_INET6, socket.SOCK_STREAM)
            if error_count > MAX_LOG_IN_ROW and self.verbose:
                logger.debug("Surpressed %d messages for failed connection attempts", error_count - MAX_LOG_IN_ROW)

        else:
            if self.verbose:
                logger.info("Connecting DLTClient using UDP Connection")

            connected = dltlib.dlt_client_connect(ctypes.byref(self), self.verbose)
            self._set_udp_multicast_buffer_size()

        if self.verbose:
            logger.info("DLT Connection return: %s", connected)
        return connected == DLT_RETURN_OK

    def disconnect(self):
        """Close all sockets"""
        if dltlib.dlt_client_cleanup(ctypes.byref(self), self.verbose) == DLT_RETURN_ERROR:
            raise RuntimeError("Could not cleanup DLTClient")
        if self._connected_socket:
            try:
                self._connected_socket.shutdown(socket.SHUT_RDWR)
            except IOError:
                pass
            except Exception:  # pylint: disable=broad-except
                logger.exception("Unexpected exception while shutting down connection")
            try:
                self._connected_socket.close()
            except IOError:
                pass
            except Exception:  # pylint: disable=broad-except
                logger.exception("Unexpected exception while disconnecting")

    def read_message(self, verbose=False):
        """Read new message

        :param bool verbose: Log every dlt_message_read(). Set True only for debugging.
        :returns: A new DLTMessage on successful read, None otherwise
        :rtype: DLTMessage|None
        """
        msg = DLTMessage(verbose=verbose)
        res = dltlib.dlt_message_read(
            ctypes.byref(msg),
            ctypes.cast(self.receiver.buf, ctypes.POINTER(ctypes.c_uint8)),
            ctypes.c_uint(self.receiver.bytesRcvd),  # length
            ctypes.c_int(0),  # resync
            ctypes.c_int(verbose),
        )  # verbose

        if res != DLT_MESSAGE_ERROR_OK:
            # - failed to read a complete message, possibly read an incomplete
            # message
            return None

        # prepare storage header
        if msg.standardheader.htyp & DLT_HTYP_WEID:
            dltlib.dlt_set_storageheader(msg.p_storageheader, msg.headerextra.ecu)
        else:
            dltlib.dlt_set_storageheader(msg.p_storageheader, ctypes.c_char_p(b""))

        return msg

    # NEW_API - ensure backwards compatibility
    @property
    def serial_mode(self):
        """Get the mode"""
        return getattr(self, "mode", getattr(super(DLTClient, self), "serial_mode", 0))

    @ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(DLTMessage), ctypes.c_void_p)
    def msg_callback(msg, data):  # pylint: disable=no-self-argument
        """Implements a simple callback that prints a dlt message received"""
        if msg is None:
            print("NULL message in callback")
            return -1
        if msg.contents.p_standardheader.contents.htyp & DLT_HTYP_WEID:
            dltlib.dlt_set_storageheader(msg.contents.p_storageheader, msg.contents.headerextra.ecu)
        else:
            dltlib.dlt_set_storageheader(msg.contents.p_storageheader, ctypes.c_char_p(b""))

        print(msg.contents)
        return 0

    def client_loop(self):
        """Executes native dlt_client_main_loop() after registering msg_callback method as callback"""
        dltlib.dlt_client_register_message_callback(self.msg_callback)
        dltlib.dlt_client_main_loop(ctypes.byref(self), None, self.verbose)

    def _set_udp_multicast_buffer_size(self, custom_fd_buffer_size_bytes=None, custom_buffer_size_bytes=None) -> None:
        fd_buffer_size = int(self._udp_fd_buffer_size_bytes or custom_fd_buffer_size_bytes or 0)
        buffer_size_bytes = int(self._udp_buffer_size_bytes or custom_buffer_size_bytes or 0)

        if fd_buffer_size:
            # Socket options are associated with an open file description. This
            # means that file descriptors duplicated as a consequence of dup()
            # (or similar) or fork() share the same set of socket options.
            #     -- Chapter 61.9 Socket Options.
            #        The Linux Programming Interface, p.1279
            #
            # The buffer size can be changed with a new fd which is created by
            # dup system call (it's the internal implementation in
            # `socket.fromfd`), so the code creates a socket instance first
            # configures  it and directly close it.
            with socket.fromfd(self.sock, socket.AF_INET, socket.SOCK_DGRAM) as conf_socket:
                logger.debug("Set UDP Multicast socket buffer size: %s kbytes", fd_buffer_size / 1024)
                conf_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, fd_buffer_size)

                real_buffer_size = int(conf_socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF) / 2)
                if real_buffer_size != fd_buffer_size:
                    logger.warning(
                        (
                            "Failed to set UDP Multicast buffer size. set_size: %s, real_size: %s. "
                            "Bypass the error and continue"
                        ),
                        fd_buffer_size / 1024,
                        real_buffer_size / 1024,
                    )
                    logger.warning(
                        (
                            "Please run command `sysctl -w net.core.rmem_max=%s` with root permission to "
                            "set the maximum size and restart dlt again."
                        ),
                        fd_buffer_size,
                    )

        if buffer_size_bytes:
            logger.debug("Set UDP Multicast DltReceiver buffer size: %s kbytes", buffer_size_bytes / 1024)
            ret = dltlib.dlt_receiver_init(
                ctypes.byref(self.receiver), self.sock, self.receiver.type, buffer_size_bytes
            )
            if ret < 0:
                raise RuntimeError(
                    f"Failed to set UDP Multicast DltReceiver buffer size. return code: {ret}, "
                    f"buffer_size_bytes: {buffer_size_bytes}"
                )


def py_dlt_file_main_loop(dlt_reader, limit=None, callback=None):
    """Main loop to read dlt messages from dlt file."""
    try:
        for msg in dlt_reader:
            # send the message to the callback and check whether we
            # need to continue
            if callback and not callback(msg):
                logger.debug("callback returned 'False'. Stopping main loop")
                return False

            if limit:
                limit -= 1
                if limit == 0:
                    break
    except IOError as err:
        # If the dlt file is empty, main_loop should not break, so it returns True
        if str(err) == DLT_EMPTY_FILE_ERROR:
            logger.debug("Dlt file is empty now. Wait until content is written")
            return True
        raise err

    return True


# pylint: disable=too-many-arguments,too-many-return-statements,too-many-branches
def py_dlt_client_main_loop(client, limit=None, verbose=0, dumpfile=None, callback=None):
    """Reimplementation of dlt_client.c:dlt_client_main_loop() in order to handle callback
    function return value
    """
    bad_messages = 0
    while True:
        if bad_messages > 100:
            # Some bad data is coming in and we can not recover - raise an error to cause a reconnect
            logger.warning("Dropping connection due to multiple malformed messages")
            return False
        # check connection status by peeking on the socket for data.
        # Note that if the remote connection is abruptly terminated,
        # this will raise a socket.timeout exception which the caller is
        # expected to handle (possibly by attempting a reconnect)
        # pylint: disable=protected-access
        if not client.is_udp_multicast:
            try:
                ready_to_read = client._connected_socket.recv(1, socket.MSG_PEEK | socket.MSG_DONTWAIT)
            except OSError as os_exc:
                logger.error("[%s]: DLTLib closed connected socket", os_exc)
                return False

            if not ready_to_read:
                # - implies that the other end has called close()/shutdown()
                # (ie: clean disconnect)
                logger.debug("connection terminated, returning")
                return False

        # - check if stop flag has been set (end of loop)
        if callback and not callback(None):
            logger.debug("callback returned 'False'. Stopping main loop")
            return False

        # we now have data to read. Note that dlt_receiver_receive()
        # is a blocking call that only returns if there is data to be
        # read or the remote end closes connection. So, irrespective of
        # the status of the callback (in the case of dlt_broker, this is
        # the stop_flag Event), this loop will only proceed after the
        # function has returned or terminate when an exception is raised
        if hasattr(client.receiver, "type"):
            recv_size = dltlib.dlt_receiver_receive(ctypes.byref(client.receiver))
        else:
            recv_size = dltlib.dlt_receiver_receive(ctypes.byref(client.receiver), DLT_RECEIVE_SOCKET)
        if recv_size <= 0:
            logger.error("Error while reading from socket")
            return False

        msg = client.read_message(verbose)
        while msg:
            try:
                if msg.apid == b"" and msg.ctid == b"":
                    logger.debug("Received a corrupt message")
                    bad_messages += 1
            except AttributeError:
                logger.debug("Skipping a very corrupted message")
                bad_messages += 1
                msg = client.read_message()
                continue

            bad_messages = 0
            # save the message
            if dumpfile:
                dumpfile.write(msg.to_bytes())

            # remove message from receiver buffer
            size = msg.headersize + msg.datasize - ctypes.sizeof(cDltStorageHeader)
            if msg.found_serialheader:
                size += DLT_ID_SIZE

            if dltlib.dlt_receiver_remove(ctypes.byref(client.receiver), size) < 0:
                logger.error("dlt_receiver_remove failed")
                return False

            # send the message to the callback and check whether we
            # need to continue
            if callback and not callback(msg):
                logger.debug("callback returned 'False'. Stopping main loop")
                break

            if limit is not None:
                limit -= 1
                if limit == 0:
                    break

            # read the next message
            msg = client.read_message()
        else:
            # - failed to read a complete message, rewind the client
            # receiver buffer pointer to start of the buffer
            if dltlib.dlt_receiver_move_to_begin(ctypes.byref(client.receiver)) == DLT_RETURN_ERROR:
                logger.error("dlt_receiver_move_to_begin failed")
                return False

        # Check if we need to keep going
        if callback and not callback(msg):
            logger.debug("callback returned 'False'. Stopping main loop")
            break

    return True


def save(messages, filename, append=False):
    """Save DLT messages to a file

    :param list messages: List of messages to save
    :param str filename: Filename for the DLT log file the messages will be stored to
    :param bool append: New data will be appended to an existing file if set to True
    """
    with open(filename, "ab" if append else "wb") as tracefile:
        for msg in messages:
            tracefile.write(msg.to_bytes())


def load(filename, filters=None, split=False, verbose=False, live_run=False):
    """Load DLT messages from a file

    :param str filename: Filename for the DLT log file the messages will be store to
    :param list filters: List of filters to apply [("APPID", "CTID"), ...]
    :param bool split: Ignored - compatibility option
    :param bool verbose: Be verbose
    :returns: A DLTFile object
    :rtype: DLTFile object
    """
    cfile = cDLTFile(filename=filename, is_live=live_run)
    cfile.set_filters(filters)
    return cfile
